"""Unit tests for serialized_sections module."""
from pathlib import Path

import pytest

from bluecellulab import Cell
from bluecellulab.cell.serialized_sections import SerializedSections


script_dir = Path(__file__).parent.parent


@pytest.mark.v5
def test_serialized_sections():
    """Test the SerializedSections class."""
    cell = Cell(
        "%s/examples/cell_example1/test_cell.hoc" % script_dir,
        "%s/examples/cell_example1" % script_dir)

    serialized_sections = SerializedSections(cell.cell.getCell())
    assert len(serialized_sections.isec2sec) == 135
    assert serialized_sections.isec2sec[0].nchild() == 7
    assert serialized_sections.isec2sec[0].has_parent() is False
    assert serialized_sections.isec2sec[0].has_trueparent() is False
    for sec in serialized_sections.isec2sec.values():
        assert sec.exists()

    # check edge cases
    modified_cell = cell.cell.getCell()

    # get the first section of modified_cell
    section_list = modified_cell.all
    first_section = next(iter(section_list))
    first_section(0.0001).v = -2
    with pytest.warns(UserWarning):
        SerializedSections(modified_cell)

    modified_cell.nSecAll = -1
    with pytest.raises(ValueError):
        SerializedSections(modified_cell)



"""Unit tests for tools.py"""

from pathlib import Path

import numpy as np
import pytest

import bluecellulab
from bluecellulab.cell.ballstick import create_ball_stick
from bluecellulab.circuit.circuit_access import EmodelProperties
from bluecellulab.tools import template_accepts_cvode, check_empty_topology

script_dir = Path(__file__).parent


@pytest.mark.v5
def test_calculate_SS_voltage_subprocess():
    """Tools: Test calculate_SS_voltage"""
    SS_voltage = bluecellulab.calculate_SS_voltage_subprocess(
        template_path=script_dir / "examples/cell_example1/test_cell.hoc",
        morphology_path=script_dir / "examples/cell_example1",
        template_format="v5",
        emodel_properties=None,
        step_level=0)
    assert abs(SS_voltage - -73.9235504304) < 0.001

    SS_voltage_stoch = bluecellulab.calculate_SS_voltage_subprocess(
        template_path=script_dir / "examples/cell_example2/test_cell.hoc",
        morphology_path=script_dir / "examples/cell_example2",
        template_format="v5",
        emodel_properties=None,
        step_level=0)
    assert abs(SS_voltage_stoch - -73.9235504304) < 0.001


def test_detect_spike():
    """Unit test for detect_spike."""

    # Case where there is a spike
    voltage_with_spike = np.array([-80, -70, -50, -10, -60, -80])
    assert bluecellulab.detect_spike(voltage_with_spike) is True

    # Case where there is no spike
    voltage_without_spike = np.array([-80, -70, -60, -50, -60, -80])
    assert bluecellulab.detect_spike(voltage_without_spike) is False

    # Edge case where the voltage reaches exactly -20 mV but does not surpass it
    voltage_at_edge = np.array([-80, -70, -60, -20, -60, -80])
    assert bluecellulab.detect_spike(voltage_at_edge) is False

    # Test with an empty array
    voltage_empty = np.array([])
    assert bluecellulab.detect_spike(voltage_empty) is False


@pytest.mark.v6
class TestOnSonataCell:
    def setup_method(self):
        self.template_name = (
            script_dir
            / "examples/circuit_sonata_quick_scx/components/hoc/cADpyr_L2TPC.hoc"
        )
        self.morphology_path = (
            script_dir
            / "examples/circuit_sonata_quick_scx/components/morphologies/asc/rr110330_C3_idA.asc"
        )
        self.template_format = "v6"
        self.emodel_properties = EmodelProperties(
            threshold_current=0.03203125, holding_current=-0.11, AIS_scaler=1.11
        )

    def test_detect_hyp_current(self):
        """Unit test detect_hyp_current."""
        target_voltage = -85
        hyp_current = bluecellulab.detect_hyp_current(
            template_path=self.template_name,
            morphology_path=self.morphology_path,
            template_format=self.template_format,
            emodel_properties=self.emodel_properties,
            target_voltage=target_voltage,
        )
        assert hyp_current == pytest.approx(-0.0009765625)

    def test_calculate_input_resistance(self):
        """Unit test calculate_input_resistance."""
        input_resistance = bluecellulab.calculate_input_resistance(
            template_path=self.template_name,
            morphology_path=self.morphology_path,
            template_format=self.template_format,
            emodel_properties=self.emodel_properties,
        )

        assert input_resistance == pytest.approx(375.88, abs=0.02)

    def test_calculate_SS_voltage(self):
        """Unit test calculate_SS_voltage."""
        step_level = 0
        SS_voltage = bluecellulab.calculate_SS_voltage(
            template_path=self.template_name,
            morphology_path=self.morphology_path,
            template_format=self.template_format,
            emodel_properties=self.emodel_properties,
            step_level=step_level,
        )
        assert SS_voltage == pytest.approx(-85.13, abs=0.02)

    def test_calculate_SS_voltage_subprocess_exception(self):
        """Unit test calculate_SS_voltage_subprocess."""
        step_level = 2
        with pytest.raises(bluecellulab.UnsteadyCellError):
            _ = bluecellulab.calculate_SS_voltage_subprocess(
                template_path=self.template_name,
                morphology_path=self.morphology_path,
                template_format=self.template_format,
                emodel_properties=self.emodel_properties,
                step_level=step_level,
                check_for_spiking=True,
            )

    def test_template_accepts_cvode(self):
        """Unit test for template_accepts_cvode."""
        assert template_accepts_cvode(self.template_name) is True

        mock_template_with_stochkv = script_dir / "examples/cell_example2/test_cell.hoc"
        assert template_accepts_cvode(mock_template_with_stochkv) is False

    def test_detect_spike_step(self):
        """Unit test for detect_spike_step."""
        hyp_level = -2
        inj_start = 100  # some start time for the current injection
        inj_stop = 200  # some stop time for the current injection
        step_level = 4  # some current level for the step

        spike_occurred = bluecellulab.detect_spike_step(
            template_path=self.template_name,
            morphology_path=self.morphology_path,
            template_format=self.template_format,
            emodel_properties=self.emodel_properties,
            hyp_level=hyp_level,
            inj_start=inj_start,
            inj_stop=inj_stop,
            step_level=step_level,
        )
        assert spike_occurred is True

    def test_search_threshold_current(self):
        """Unit test for search_threshold_current."""
        hyp_level = -2
        inj_start, inj_stop = 100, 200
        min_current, max_current = 0, 10

        threshold_current = bluecellulab.search_threshold_current(
            template_name=self.template_name,
            morphology_path=self.morphology_path,
            template_format=self.template_format,
            emodel_properties=self.emodel_properties,
            hyp_level=hyp_level,
            inj_start=inj_start,
            inj_stop=inj_stop,
            min_current=min_current,
            max_current=max_current,
        )
        assert threshold_current == pytest.approx(2.00195, abs=1e-5)

    def test_detect_spike_step_subprocess(self):
        """Unit test for detect_spike_step_subprocess."""
        hyp_level = -2
        inj_start = 100
        inj_stop = 200
        step_level = 4

        spike_occurred = bluecellulab.detect_spike_step_subprocess(
            template_path=self.template_name,
            morphology_path=self.morphology_path,
            template_format=self.template_format,
            emodel_properties=self.emodel_properties,
            hyp_level=hyp_level,
            inj_start=inj_start,
            inj_stop=inj_stop,
            step_level=step_level,
        )
        assert spike_occurred is True


@pytest.mark.v6
class TestOnSonataCircuit:
    def setup_method(self):
        self.circuit_path = (
            script_dir
            / "examples/sim_quick_scx_sonata_multicircuit/simulation_config_noinput.json"
        )
        self.cell_id = ("NodeA", 0)

    def test_holding_current(self):
        """Unit test for holding_current on a SONATA circuit."""
        v_hold = -70  # arbitrary holding voltage for the test
        i_hold, v_control = bluecellulab.holding_current(
            v_hold, cell_id=self.cell_id, circuit_path=self.circuit_path
        )

        assert i_hold == pytest.approx(0.0020779234)
        assert v_control == pytest.approx(v_hold)

    def test_holding_current_subprocess(self):
        """Unit test for holding_current_subprocess on a SONATA circuit."""
        v_hold = -80
        ssim = bluecellulab.SSim(self.circuit_path)
        cell_id = bluecellulab.circuit.node_id.create_cell_id(self.cell_id)
        cell_kwargs = ssim.fetch_cell_kwargs(cell_id)
        i_hold, v_control = bluecellulab.holding_current_subprocess(
            v_hold, enable_ttx=True, cell_kwargs=cell_kwargs
        )
        assert i_hold == pytest.approx(-0.03160848349)
        assert v_control == pytest.approx(v_hold)


def test_check_empty_topology():
    """Unit test for the check_empty_topology function."""
    assert check_empty_topology() is True
    cell = create_ball_stick()
    assert check_empty_topology() is False



"""Unit tests for the stimuli_generator module."""

from pytest import approx

import neuron
from bluecellulab.cell.stimuli_generator import (
    gen_shotnoise_signal,
    get_relative_shotnoise_params,
    gen_ornstein_uhlenbeck,
)


def test_gen_shotnoise_signal():
    """Test if the shotnoise signal is generated correctly."""
    rng = neuron.h.Random()
    rng.Random123(1, 2, 3)
    time_vec, stim_vec = gen_shotnoise_signal(4.0, 0.4, 2E3, 40E-3, 16E-4, 2,
                                              rng=rng)
    assert list(time_vec) == approx([0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75, 2.0, 2.0])
    assert list(stim_vec) == approx([0.0, 0.0, 0.0, 0.0, 0.0, 0.0700357,
                                     0.1032799, 0.1170881, 0.1207344, 0.0])

    # test with None rng
    time_vec, stim_vec = gen_shotnoise_signal(4.0, 0.4, 2E3, -40E-3, 16E-4, 2,
                                              rng=None)


def test_get_relative_shotnoise_params():
    """Unit test for _get_relative_shotnoise_params."""
    rate, amp_mean, amp_var = get_relative_shotnoise_params(
        mean=40e-3, var=16e-4, tau_D=4.0, tau_R=0.4, cv_square=0.63**2
    )
    assert rate == approx(158.73863636363635)
    assert amp_mean == approx(0.048776006926722876)
    assert amp_var == approx(0.0009442643342459686)


def test_gen_ornstein_uhlenbeck():
    """Test if the ornstein uhlenbeck signal is generated correctly."""
    rng = neuron.h.Random()
    rng.Random123(1, 2, 3)
    time_vec, stim_vec = gen_ornstein_uhlenbeck(2.8, 0.0042, 0.029, 2, rng=rng)
    assert list(time_vec) == approx([0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5,
                                     1.75, 2.0, 2.0])
    assert list(stim_vec) == approx(
        [0.029, 0.02933925, 0.02959980, 0.03052109, 0.02882802, 0.03156533,
            0.03289219, 0.03357043, 0.03049419, 0.0])

    # test with None rng
    time_vec, stim_vec = gen_ornstein_uhlenbeck(1e-10, 0.0042, 0.029, 2, rng=None)



import pytest
import numpy as np

from bluecellulab.stimulus.factory import CombinedStimulus, EmptyStimulus, Stimulus, Step, Ramp, StimulusFactory


class TestStimulus:

    def test_abstract_methods(self):
        with pytest.raises(TypeError):
            s = Stimulus(0.1)

    def test_repr(self):
        s = Step(0.1, 1, 2, 3)
        assert repr(s) == "Step(dt=0.1)"

    def test_len(self):
        s = Step(0.1, 1, 2, 3)
        assert len(s) == 10

    def test_plot(self):
        s = Step(0.1, 1, 2, 3)
        ax = s.plot()
        assert ax.get_xlabel() == "Time (ms)"
        assert ax.get_ylabel() == "Current (nA)"
        assert ax.get_title() == "Step"

    def test_plot_during_simulation(self):
        s = Step(0.1, 1, 2, 3)
        duration = 15
        ax = s.plot_during_simulation(duration)
        assert ax.get_xlim() == (0, duration)


class TestStimulusFactory:

    def setup_method(self):
        self.dt = 0.1
        self.factory = StimulusFactory(dt=self.dt)

    def test_create_step(self):
        s = self.factory.step(1, 2, 3)
        assert isinstance(s, Step)
        assert np.all(s.time == np.arange(1, 2, self.dt))
        assert np.all(s.current == np.full(10, 3))

    def test_create_ramp(self):
        s = self.factory.ramp(1, 2, 3, 4)
        assert isinstance(s, Ramp)
        assert np.all(s.time == np.arange(1, 2, self.dt))
        assert np.all(s.current == np.linspace(3, 4, 10))

    def test_create_ap_waveform(self):
        s = self.factory.ap_waveform(threshold_current=1)
        assert isinstance(s, CombinedStimulus)
        # below write np almost equal
        np.testing.assert_allclose(s.time, np.arange(250, 550, self.dt))
        assert s.current[0] == 2.2
        assert s.current[-1] == 0.0

    def test_create_idrest(self):
        s = self.factory.idrest(threshold_current=1)
        assert isinstance(s, CombinedStimulus)
        assert s.stimulus_time == 1600

    def test_create_iv(self):
        s = self.factory.iv(threshold_current=1)
        assert isinstance(s, CombinedStimulus)
        assert s.stimulus_time == 3250
        # assert there are negative values
        assert np.any(s.current < 0)
        # assert no positive values
        assert not np.any(s.current > 0)

    def test_create_fire_pattern(self):
        s = self.factory.fire_pattern(threshold_current=1)
        assert isinstance(s, CombinedStimulus)
        assert s.stimulus_time == 3850


def test_combined_stimulus():
    """Test combining Stimulus objects."""
    s1 = Step(0.1, 1, 2, 3)
    s2 = Step(0.1, 3, 4, 5)
    combined = s1 + s2

    assert isinstance(combined, CombinedStimulus)
    assert np.all(combined.time == np.concatenate([s1.time, s2.time + s1.time[-1] + 0.1]))
    assert np.all(combined.current == np.concatenate([s1.current, s2.current]))
    assert combined.dt == 0.1
    assert len(combined) == len(s1) + len(s2)
    assert combined.stimulus_time == (len(combined) * combined.dt)


def test_combined_stimulus_different_dt():
    s1 = Step(0.1, 1, 2, 3)
    s2 = Step(0.2, 3, 4, 5)
    with pytest.raises(ValueError):
        combined = s1 + s2


def test_empty_stimulus():
    dt = 0.2
    duration = 1.0
    stimulus = EmptyStimulus(dt, duration)

    assert stimulus.dt == dt
    assert np.all(stimulus.time == np.arange(0, duration, dt))
    assert np.all(stimulus.current == np.zeros_like(stimulus.time))


def test_combine_multiple_stimuli():
    """Test combining multiple stimuli."""
    dt = 0.1
    stim1 = Step(dt, 99, 101, 3)
    stim2 = Ramp(dt, 3, 4, 2, 4)
    stim3 = EmptyStimulus(dt, 1)
    stim4 = Step(dt, 5, 6, 1)

    combined = stim1 + stim2 + stim3 + stim4

    assert isinstance(combined, CombinedStimulus)
    assert combined.dt == dt

    shifted_stim2_time = stim2.time + stim1.time[-1] + dt
    shifted_stim3_time = stim3.time + shifted_stim2_time[-1] + dt
    shifted_stim4_time = stim4.time + shifted_stim3_time[-1] + dt

    expected_time = np.concatenate([
        stim1.time,
        shifted_stim2_time,
        shifted_stim3_time,
        shifted_stim4_time,
    ])
    expected_current = np.concatenate([
        stim1.current,
        stim2.current,
        stim3.current,
        stim4.current,
    ])

    assert np.all(combined.time == expected_time)
    assert np.all(combined.current == expected_current)



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dictionary of cells that support the old way of retrieval by int ids.

as well as the new way of retrieval by CellId.
"""

from __future__ import annotations

import bluecellulab
from bluecellulab.circuit.node_id import CellId, create_cell_id


class CellDict(dict):
    """Dictionary of Cell objects, backwards compatible with old gid
    representation."""

    def __setitem__(self, key: int | tuple[str, int], value: bluecellulab.Cell) -> None:
        super().__setitem__(create_cell_id(key), value)

    def __getitem__(self, key: int | tuple[str, int]) -> bluecellulab.Cell:
        """Can retrieve for single population circuits, supports cells[gid]."""
        cell_id: CellId = create_cell_id(key)
        return super().__getitem__(cell_id)



"""Unit tests for the exceptions module."""

from bluecellulab.exceptions import error_context


def test_error_context():
    """Unit test for the error_context function."""
    attr_err_msg = "hoc.HocObject' object has no attribute 'minis_single_vesicle_"
    lookup_err_msg = "'X' is not a defined hoc variable name"
    context_info = "mechanism/s for minis_single_vesicle need to be compiled"
    try:
        with error_context(context_info):
            raise AttributeError(attr_err_msg)
    except AttributeError as error:
        assert str(error) == f"{context_info}: {attr_err_msg}"
    try:
        with error_context(context_info):
            raise LookupError(lookup_err_msg)
    except LookupError as error:
        assert str(error) == f"{context_info}: {lookup_err_msg}"



"""Unit tests for TTX in mod files"""

import os

import pytest

import bluecellulab

script_dir = os.path.dirname(__file__)


@pytest.mark.v5
def test_allNaChannels():
    """TTX: Testing ttx enabling"""

    na_channelnames = ['NaTs2_t']

    cell = bluecellulab.Cell(
        "%s/examples/cell_example_empty/test_cell.hoc" %
        script_dir,
        "%s/examples/cell_example_empty" %
        script_dir)

    for na_channelname in na_channelnames:
        cell.soma.insert(na_channelname)

        cell.add_step(0, 1000, .1)
        sim = bluecellulab.Simulation()
        sim.add_cell(cell)

        sim.run(10)
        voltage_nottx1 = cell.get_soma_voltage()

        cell.enable_ttx()
        sim.run(10)
        voltage_ttx = cell.get_soma_voltage()

        cell.disable_ttx()
        sim.run(10)
        voltage_nottx2 = cell.get_soma_voltage()

        # Check if voltage changed due to enable_ttx
        assert voltage_nottx1[-1] != voltage_ttx[-1]

        assert voltage_nottx1[-1] == voltage_nottx2[-1]


@pytest.mark.v6
def test_allNaChannels_v6a():
    """TTX: Testing ttx enabling in v6a cell"""

    cell = bluecellulab.Cell(
        "%s/examples/cell_example_empty/test_cell_v6a.hoc" %
        script_dir,
        "%s/examples/cell_example_empty" %
        script_dir)

    cell.add_step(0, 1000, .1)
    sim = bluecellulab.Simulation()
    sim.add_cell(cell)

    sim.run(10)
    # time_nottx1 = cell.get_time()
    voltage_nottx1 = cell.get_soma_voltage()

    cell.enable_ttx()
    sim.run(10)
    # time_ttx = cell.get_time()
    voltage_ttx = cell.get_soma_voltage()

    cell.disable_ttx()
    sim.run(10)
    voltage_nottx2 = cell.get_soma_voltage()

    '''
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    plt.plot(time_nottx1, voltage_nottx1, label='no ttx')
    plt.plot(time_ttx, voltage_ttx, label='ttx')
    plt.legend()
    plt.savefig('ttx.png')
    '''

    # Check if voltage changed due to enable_ttx

    assert voltage_nottx1[-1] != voltage_ttx[-1]
    assert voltage_nottx1[-1] == voltage_nottx2[-1]



from __future__ import annotations
from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import numpy as np


class Stimulus(ABC):
    def __init__(self, dt: float) -> None:
        self.dt = dt

    @property
    @abstractmethod
    def time(self) -> np.ndarray:
        """Time values of the stimulus."""
        ...

    @property
    @abstractmethod
    def current(self) -> np.ndarray:
        """Current values of the stimulus."""
        ...

    def __len__(self) -> int:
        return len(self.time)

    @property
    def stimulus_time(self) -> float:
        return len(self) * self.dt

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dt={self.dt})"

    def plot(self, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        ax.plot(self.time, self.current, **kwargs)
        ax.set_xlabel("Time (ms)")
        ax.set_ylabel("Current (nA)")
        ax.set_title(self.__class__.__name__)
        return ax

    def plot_during_simulation(self, duration: float, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        # Create an array for the entire duration
        full_time = np.arange(0, duration, self.dt)
        full_current = np.zeros_like(full_time)

        # Replace the corresponding values with self.time and self.current
        indices = (self.time / self.dt).astype(int)
        full_current[indices] = self.current

        ax.plot(full_time, full_current, **kwargs)
        ax.set_xlabel("Time (ms)")
        ax.set_ylabel("Current (nA)")
        ax.set_title(self.__class__.__name__)
        ax.set_xlim(0, duration)
        return ax

    def __add__(self, other: Stimulus) -> CombinedStimulus:
        """Override + operator to concatenate Stimulus objects."""
        if self.dt != other.dt:
            raise ValueError("Stimulus objects must have the same dt to be concatenated")
        # shift other time
        other_time = other.time + self.time[-1] + self.dt
        combined_time = np.concatenate([self.time, other_time])
        # Concatenate the current arrays
        combined_current = np.concatenate([self.current, other.current])
        return CombinedStimulus(self.dt, combined_time, combined_current)


class CombinedStimulus(Stimulus):
    """Represents the Stimulus created by combining multiple stimuli."""
    def __init__(self, dt: float, time: np.ndarray, current: np.ndarray) -> None:
        super().__init__(dt)
        self._time = time
        self._current = current

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current


class EmptyStimulus(Stimulus):
    """Represents empty stimulus (all zeros) that has no impact on the cell.

    This is required by some Stimuli that expect the cell to rest.
    """
    def __init__(self, dt: float, duration: float) -> None:
        super().__init__(dt)
        self.duration = duration

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.zeros_like(self.time)


class Step(Stimulus):
    def __init__(self, dt: float, start: float, end: float, amplitude: float) -> None:
        super().__init__(dt)
        self.start = start
        self.end = end
        self.amplitude = amplitude

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        threshold_current: float,
        threshold_percentage: float,
        start: float,
        end: float,
        post_wait: float,
    ) -> CombinedStimulus:
        amplitude = threshold_current * threshold_percentage / 100
        res = cls(dt, start, end, amplitude) + EmptyStimulus(dt, duration=post_wait)
        return res

    @property
    def time(self) -> np.ndarray:
        return np.arange(self.start, self.end, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.full_like(self.time, self.amplitude)


class Ramp(Stimulus):
    def __init__(
        self,
        dt: float,
        start: float,
        end: float,
        amplitude_start: float,
        amplitude_end: float,
    ) -> None:
        super().__init__(dt)
        self.start = start
        self.end = end
        self.amplitude_start = amplitude_start
        self.amplitude_end = amplitude_end

    @property
    def time(self) -> np.ndarray:
        return np.arange(self.start, self.end, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.linspace(self.amplitude_start, self.amplitude_end, len(self.time))


class StimulusFactory:
    def __init__(self, dt: float):
        self.dt = dt

    def step(self, start: float, end: float, amplitude: float) -> Stimulus:
        return Step(self.dt, start, end, amplitude)

    def ramp(
        self, start: float, end: float, amplitude_start: float, amplitude_end: float
    ) -> Stimulus:
        return Ramp(self.dt, start, end, amplitude_start, amplitude_end)

    def ap_waveform(
        self,
        threshold_current: float,
        threshold_percentage: float = 220.0,
        start: float = 250.0,
        end: float = 300.0,
        post_wait: float = 250.0,
    ) -> Stimulus:
        """Returns the APWaveform Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            start: The start time of the step.
            end: The end time of the step.
            post_wait: The time to wait after the end of the step.
        """
        return Step.threshold_based(
            self.dt, threshold_current, threshold_percentage, start, end, post_wait
        )

    def idrest(
        self,
        threshold_current: float,
        threshold_percentage: float = 200.0,
        start: float = 250.0,
        end: float = 1600.0,
        post_wait: float = 250.0,
    ) -> Stimulus:
        """Returns the IDRest Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            start: The start time of the step.
            end: The end time of the step.
            post_wait: The time to wait after the end of the step.
        """
        return Step.threshold_based(
            self.dt, threshold_current, threshold_percentage, start, end, post_wait
        )

    def iv(
        self,
        threshold_current: float,
        threshold_percentage: float = -40.0,
        start: float = 250.0,
        end: float = 3250.0,
        post_wait: float = 250.0,
    ) -> Stimulus:
        """Returns the IV Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            start: The start time of the step.
            end: The end time of the step.
            post_wait: The time to wait after the end of the step.
        """
        return Step.threshold_based(
            self.dt, threshold_current, threshold_percentage, start, end, post_wait
        )

    def fire_pattern(
        self,
        threshold_current: float,
        threshold_percentage: float = 200.0,
        start: float = 250.0,
        end: float = 3850.0,
        post_wait: float = 250.0,
    ) -> Stimulus:
        """Returns the FirePattern Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            start: The start time of the step.
            end: The end time of the step.
            post_wait: The time to wait after the end of the step.
        """
        return Step.threshold_based(
            self.dt, threshold_current, threshold_percentage, start, end, post_wait
        )



"""Unit tests for the cell_dict module."""
from pathlib import Path

import pytest

from bluecellulab.cell import Cell, CellDict
from bluecellulab.circuit.node_id import CellId


parent_dir = Path(__file__).resolve().parent.parent


@pytest.mark.v5
def test_cell_dict():
    """Unit test for the CellDict class."""
    cell = Cell(
        "%s/examples/cell_example1/test_cell_longname1.hoc" % str(parent_dir),
        "%s/examples/cell_example1" % str(parent_dir))

    cell_dict = CellDict()
    cell_dict[1] = cell
    assert cell_dict[("", 1)] == cell
    assert cell_dict[1] == cell

    cell_dict[("population1", 2)] = cell
    cell_id = CellId("population1", 2)
    assert cell_dict[cell_id] == cell
    assert cell_dict[("population1", 2)] == cell



"""Unit tests for cell.sonata_proxy."""
from pathlib import Path

from bluecellulab.cell import SonataProxy
from bluecellulab.circuit import SonataCircuitAccess
from bluecellulab.circuit.node_id import CellId
from bluecellulab.exceptions import MissingSonataPropertyError

from pytest import approx, raises

parent_dir = Path(__file__).resolve().parent.parent


hipp_circuit_with_projections = (
    parent_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "projections"
    / "simulation_config.json"
)


class TestSonataProxy:

    def setup_method(self):
        circuit_access = SonataCircuitAccess(str(hipp_circuit_with_projections))
        cell_id = CellId("hippocampus_neurons", 1)
        self.sonata_proxy = SonataProxy(cell_id, circuit_access)

    def test_get_property(self):
        holding_current = self.sonata_proxy.get_property("@dynamics:holding_current")
        assert holding_current.iloc[0] == approx(-0.116351)

        threshold_current = self.sonata_proxy.get_property("@dynamics:threshold_current")
        assert threshold_current.iloc[0] == approx(0.332031)

    def test_get_input_resistance(self):
        with raises(MissingSonataPropertyError):
            self.sonata_proxy.get_input_resistance()



"""Unit tests for Cell.py"""

import math
import random
import warnings
from pathlib import Path

import neuron
import numpy as np
import pytest

import bluecellulab
from bluecellulab.circuit.circuit_access import EmodelProperties
from bluecellulab.cell.template import NeuronTemplate, shorten_and_hash_string
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.ssim import SSim

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ndarray size changed")

parent_dir = Path(__file__).resolve().parent.parent


@pytest.mark.v5
def test_longname():
    """Cell: Test loading cell with long name"""

    cell = bluecellulab.Cell(
        "%s/examples/cell_example1/test_cell_longname1.hoc" % str(parent_dir),
        "%s/examples/cell_example1" % str(parent_dir))
    assert isinstance(cell, bluecellulab.Cell)

    del cell


@pytest.mark.v5
def test_load_template():
    """Test the neuron template loading."""
    hoc_path = parent_dir / "examples/cell_example1/test_cell.hoc"
    morph_path = parent_dir / "examples/cell_example1/test_cell.asc"
    template = NeuronTemplate(hoc_path, morph_path, "v5", None)
    template_name = template.template_name
    assert template_name == f"test_cell_bluecellulab_{hex(id(template))}"


def test_shorten_and_hash_string():
    """Unit test for the shorten_and_hash_string function."""
    with pytest.raises(ValueError):
        shorten_and_hash_string(label="1", hash_length=21)

    short_label = "short-label"
    assert shorten_and_hash_string(short_label) == short_label

    long_label = "test-cell" * 10
    assert len(shorten_and_hash_string(long_label)) < len(long_label)


@pytest.mark.v5
class TestCellBaseClass1:
    """First Cell test class"""

    def setup_method(self):
        """Setup"""
        self.cell = bluecellulab.Cell(
            "%s/examples/cell_example1/test_cell.hoc" % str(parent_dir),
            "%s/examples/cell_example1" % str(parent_dir))
        assert isinstance(self.cell, bluecellulab.Cell)

    def teardown_method(self):
        """Teardown"""
        del self.cell

    def test_fields(self):
        """Cell: Test the fields of a Cell object"""
        assert isinstance(self.cell.soma, neuron.nrn.Section)
        assert isinstance(self.cell.axonal[0], neuron.nrn.Section)
        assert math.fabs(self.cell.threshold - 0.184062) < 0.00001
        assert math.fabs(self.cell.hypamp - -0.070557) < 0.00001
        # Lowered precision because of commit
        # 81a7a398214f2f5fba199ac3672c3dc3ccb6b103
        # in nrn simulator repo
        assert math.fabs(self.cell.soma.diam - 13.78082) < 0.0001
        assert math.fabs(self.cell.soma.L - 19.21902) < 0.00001
        assert math.fabs(self.cell.basal[2].diam - 0.595686) < 0.00001
        assert math.fabs(self.cell.basal[2].L - 178.96164) < 0.00001
        assert math.fabs(self.cell.apical[10].diam - 0.95999) < 0.00001
        assert math.fabs(self.cell.apical[10].L - 23.73195) < 0.00001

    def test_get_psection(self):
        """Cell: Test cell.get_psection"""
        idx = 0
        name = "Cell[0].soma[0]"
        assert isinstance(
            self.cell.get_psection(idx).hsection, neuron.nrn.Section)
        assert isinstance(
            self.cell.get_psection(name).hsection, neuron.nrn.Section)
        assert self.cell.get_psection(idx) == self.cell.get_psection(name)
        with pytest.raises(BluecellulabError):
            self.cell.get_psection(None)
        with pytest.raises(BluecellulabError):
            self.cell.get_psection(5.8673453123)

    def test_add_recording(self):
        """Cell: Test cell.add_recording"""
        varname = 'self.apical[1](0.5)._ref_v'
        self.cell.add_recording(varname)
        assert varname in self.cell.recordings

        second_varname = 'self.apical[1](0.6)._ref_v'
        self.cell.add_recording(second_varname, dt=0.025)
        assert second_varname in self.cell.recordings

    def test_add_recordings(self):
        """Cell: Test cell.add_recordings"""
        varnames = [
            'self.axonal[0](0.25)._ref_v',
            'self.soma(0.5)._ref_v',
            'self.apical[1](0.5)._ref_v']
        self.cell.add_recordings(varnames)
        for varname in varnames:
            assert varname in self.cell.recordings

    def test_add_ais_recording(self):
        """Cell Test add_ais_recording."""
        self.cell.add_ais_recording()
        ais_key = "self.axonal[1](0.5)._ref_v"
        assert ais_key in self.cell.recordings

    def test_add_allsections_voltagerecordings(self):
        """Cell: Test cell.add_allsections_voltagerecordings"""
        self.cell.add_allsections_voltagerecordings()

        all_sections = self.cell.cell.getCell().all
        for section in all_sections:
            varname = 'neuron.h.%s(0.5)._ref_v' % section.name()
            assert varname in self.cell.recordings

    def test_manual_add_allsection_voltage_recordings(self):
        """Cell: Test cell.add_voltage_recording."""
        all_sections = self.cell.cell.getCell().all
        last_section = None
        for section in all_sections:
            self.cell.add_voltage_recording(section, 0.5)
            recording = self.cell.get_voltage_recording(section, 0.5)
            assert len(recording) == 0
            last_section = section
        with pytest.raises(BluecellulabError):
            self.cell.get_voltage_recording(last_section, 0.7)

    def test_get_allsections_voltagerecordings(self):
        """Cell: Test cell.get_allsections_voltagerecordings."""
        self.cell.recordings.clear()

        with pytest.raises(BluecellulabError):
            recordings = self.cell.get_allsections_voltagerecordings()

        self.cell.add_allsections_voltagerecordings()
        recordings = self.cell.get_allsections_voltagerecordings()
        assert len(recordings) == len(self.cell.recordings)
        for recording in recordings:
            assert any(recording in s for s in self.cell.recordings)

    def test_euclid_section_distance(self):
        """Cell: Test cell.euclid_section_distance"""

        random.seed(1)

        location1 = 0.0
        location2 = 1.0
        for _ in range(1000):
            hsection1 = random.choice(random.choice(
                [self.cell.apical, self.cell.somatic, self.cell.basal]))
            hsection2 = random.choice(random.choice(
                [self.cell.apical, self.cell.somatic, self.cell.basal]))
            distance_euclid = \
                self.cell.euclid_section_distance(hsection1=hsection1,
                                                  hsection2=hsection2,
                                                  location1=location1,
                                                  location2=location2,
                                                  projection='xyz')
            x1 = neuron.h.x3d(0, sec=hsection1)
            y1 = neuron.h.y3d(0, sec=hsection1)
            z1 = neuron.h.z3d(0, sec=hsection1)
            x2 = neuron.h.x3d(neuron.h.n3d(sec=hsection2) - 1, sec=hsection2)
            y2 = neuron.h.y3d(neuron.h.n3d(sec=hsection2) - 1, sec=hsection2)
            z2 = neuron.h.z3d(neuron.h.n3d(sec=hsection2) - 1, sec=hsection2)

            distance_hand = np.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)
            assert distance_euclid == distance_hand


@pytest.mark.debugtest
class TestCellBaseClassVClamp:

    """First Cell test class"""

    def setup_method(self):
        """Setup"""
        self.cell = bluecellulab.Cell(
            "%s/examples/cell_example1/test_cell.hoc" % str(parent_dir),
            "%s/examples/cell_example1" % str(parent_dir))
        assert (isinstance(self.cell, bluecellulab.Cell))

    def teardown_method(self):
        """Teardown"""
        del self.cell

    def test_add_voltage_clamp(self):
        """Cell: Test add_voltage_clamp"""

        level = -90
        stop_time = 50
        total_time = 200
        rs = .1
        vclamp = self.cell.add_voltage_clamp(
            stop_time=stop_time,
            level=level,
            current_record_name='test_vclamp',
            rs=rs)

        assert vclamp.amp1 == level
        assert vclamp.dur1 == stop_time
        assert vclamp.dur2 == 0
        assert vclamp.dur3 == 0
        assert vclamp.rs == rs

        sim = bluecellulab.Simulation()
        sim.add_cell(self.cell)
        sim.run(total_time, dt=.1, cvode=False)

        time = self.cell.get_time()
        current = self.cell.get_recording('test_vclamp')
        import numpy as np

        voltage = self.cell.get_soma_voltage()

        voltage_vc_end = np.mean(
            voltage[np.where((time < stop_time) & (time > .9 * stop_time))])

        assert (abs(voltage_vc_end - level) < .1)

        voltage_end = np.mean(
            voltage
            [np.where((time < total_time) & (time > .9 * total_time))])

        assert (abs(voltage_end - (-73)) < 1)

        current_vc_end = np.mean(
            current[np.where((time < stop_time) & (time > .9 * stop_time))])

        assert (abs(current_vc_end - (-.1)) < .01)

        current_after_vc_end = np.mean(
            current[
                np.where((time > stop_time) & (time < 1.1 * stop_time))])

        assert current_after_vc_end == 0.0


class TestCellSpikes:

    def setup_method(self):
        self.cell = bluecellulab.Cell(
            f"{parent_dir}/examples/cell_example1/test_cell.hoc",
            f"{parent_dir}/examples/cell_example1")
        self.sim = bluecellulab.Simulation()
        self.sim.add_cell(self.cell)

    @pytest.mark.v5
    def test_get_recorded_spikes(self):
        """Cell: Test get_recorded_spikes."""
        self.cell.start_recording_spikes(None, "soma", -30)
        self.cell.add_step(start_time=2.0, stop_time=22.0, level=1.0)
        self.sim.run(24, cvode=False)
        spikes = self.cell.get_recorded_spikes("soma")
        ground_truth = [3.350000000100014, 11.52500000009988, 19.9750000000994]
        assert np.allclose(spikes, ground_truth)

    @pytest.mark.v5
    def test_create_netcon_spikedetector(self):
        """Cell: create_netcon_spikedetector."""
        threshold = -29.0
        netcon = self.cell.create_netcon_spikedetector(None, "AIS", -29.0)
        assert netcon.threshold == threshold
        netcon = self.cell.create_netcon_spikedetector(None, "soma", -29.0)
        with pytest.raises(ValueError):
            self.cell.create_netcon_spikedetector(None, "Dendrite", -29.0)


@pytest.mark.v6
def test_add_dendrogram():
    """Cell: Test get_recorded_spikes."""
    emodel_properties = EmodelProperties(threshold_current=1.1433533430099487,
                                         holding_current=1.4146618843078613,
                                         AIS_scaler=1.4561502933502197,
                                         soma_scaler=1.0)
    cell = bluecellulab.Cell(
        "%s/examples/circuit_sonata_quick_scx/components/hoc/cADpyr_L2TPC.hoc" % str(parent_dir),
        "%s/examples/circuit_sonata_quick_scx/components/morphologies/asc/rr110330_C3_idA.asc" % str(parent_dir),
        template_format="v6",
        emodel_properties=emodel_properties)
    cell.add_plot_window(['self.soma(0.5)._ref_v'])
    output_path = "cADpyr_L2TPC_dendrogram.png"
    cell.add_dendrogram(save_fig_path=output_path)
    sim = bluecellulab.Simulation()
    sim.add_cell(cell)
    cell.add_step(start_time=2.0, stop_time=22.0, level=1.0)
    sim.run(24, cvode=False)
    assert Path(output_path).is_file()


@pytest.mark.v6
class TestCellV6:
    """Test class for testing Cell object functionalities with v6 template."""

    def setup_method(self):
        """Setup."""
        emodel_properties = EmodelProperties(
            threshold_current=1.1433533430099487,
            holding_current=1.4146618843078613,
            AIS_scaler=1.4561502933502197,
            soma_scaler=1.0
        )
        self.cell = bluecellulab.Cell(
            "%s/examples/circuit_sonata_quick_scx/components/hoc/cADpyr_L2TPC.hoc" % str(parent_dir),
            "%s/examples/circuit_sonata_quick_scx/components/morphologies/asc/rr110330_C3_idA.asc" % str(parent_dir),
            template_format="v6",
            emodel_properties=emodel_properties
        )

    def test_repr_and_str(self):
        """Test the repr and str representations of a cell object."""
        # >>> print(cell)
        # Cell Object: <bluecellulab.cell.core.Cell object at 0x7f73b3fb2550>.
        # NEURON ID: cADpyr_L2TPC_bluecellulab_0x7f73b48e2510.
        # make sure NEURON template name is in the string representation
        assert self.cell.cell.hname().split('[')[0] in str(self.cell)

    def test_area(self):
        """Test the cell's area computation."""
        assert self.cell.area() == 5812.493415302344

    def test_cell_id(self):
        """Test for checking if cell_id is different btw. 2 cells when unspecified."""
        emodel_properties = EmodelProperties(
            threshold_current=1.1433533430099487,
            holding_current=1.4146618843078613,
            AIS_scaler=1.4561502933502197,
            soma_scaler=1.0
        )
        cell2 = bluecellulab.Cell(
            "%s/examples/circuit_sonata_quick_scx/components/hoc/cADpyr_L2TPC.hoc" % str(parent_dir),
            "%s/examples/circuit_sonata_quick_scx/components/morphologies/asc/rr110330_C3_idA.asc" % str(parent_dir),
            template_format="v6",
            emodel_properties=emodel_properties
        )
        assert self.cell.cell_id != cell2.cell_id

    def test_get_childrensections(self):
        """Test the get_childrensections method."""
        res = self.cell.get_childrensections(self.cell.soma)
        assert len(res) == 3

    def test_get_parentsection(self):
        """Test the get_parentsection method."""
        section = self.cell.get_childrensections(self.cell.soma)[0]
        res = self.cell.get_parentsection(section)
        assert res == self.cell.soma

    def test_somatic_sections(self):
        """Test that somatic property returns a non-empty list sections."""
        assert isinstance(self.cell.somatic, list)
        assert len(self.cell.somatic) > 0

    def test_basal_sections(self):
        """Test that basal property returns a non-empty list of sections."""
        assert isinstance(self.cell.basal, list)
        assert len(self.cell.basal) > 0

    def test_apical_sections(self):
        """Test that apical property returns a non-empty list sections."""
        assert isinstance(self.cell.apical, list)
        assert len(self.cell.apical) > 0

    def test_axonal_sections(self):
        """Test that axonal property returns a non-empty list of sections."""
        assert isinstance(self.cell.axonal, list)
        assert len(self.cell.axonal) > 0

    def test_all_sections(self):
        """Test that the all property returns a non-empty list of all sections."""
        assert isinstance(self.cell.all, list)
        assert len(self.cell.all) > 0


@pytest.mark.v6
def test_add_synapse_replay():
    """Cell: test add_synapse_replay."""
    sonata_sim_path = (
        parent_dir
        / "examples"
        / "sonata_unit_test_sims"
        / "synapse_replay"
        / "simulation_config.json"
    )
    ssim = bluecellulab.SSim(sonata_sim_path)
    ssim.spike_threshold = -900.0
    cell_id = ("hippocampus_neurons", 0)
    ssim.instantiate_gids(cell_id,
                          add_stimuli=True, add_synapses=True,
                          interconnect_cells=False)
    cell = ssim.cells[cell_id]
    assert len(cell.connections) == 3
    assert cell.connections[
        ("hippocampus_projections__hippocampus_neurons__chemical", 0)
    ].pre_spiketrain.tolist() == [16.0, 22.0, 48.0]


@pytest.mark.v6
class TestWithinCircuit:

    def setup_method(self):
        """Setup method called before each test method."""
        sonata_sim_path = (
            parent_dir
            / "examples"
            / "sim_quick_scx_sonata_multicircuit"
            / "simulation_config_noinput.json"
        )
        cell_id = ("NodeA", 0)
        ssim = SSim(sonata_sim_path)
        ssim.instantiate_gids(cell_id, add_synapses=True, add_stimuli=False)
        self.cell = ssim.cells[cell_id]
        self.ssim = ssim  # for persistance

    def test_pre_gids(self):
        """Test get_pre_gids within a circuit."""
        pre_gids = self.cell.pre_gids()
        assert pre_gids == [0, 1]

    def test_pre_gid_synapse_ids(self):
        """Test pre_gid_synapse_ids within a circuit."""
        pre_gids = self.cell.pre_gids()

        first_gid_synapse_ids = self.cell.pre_gid_synapse_ids(pre_gids[0])
        assert first_gid_synapse_ids == [('NodeB__NodeA__chemical', 0), ('NodeB__NodeA__chemical', 2)]

        second_gid_synapse_ids = self.cell.pre_gid_synapse_ids(pre_gids[1])
        assert len(second_gid_synapse_ids) == 4
        assert second_gid_synapse_ids[0] == ('NodeA__NodeA__chemical', 0)
        assert second_gid_synapse_ids[1] == ('NodeA__NodeA__chemical', 1)
        assert second_gid_synapse_ids[2] == ('NodeB__NodeA__chemical', 1)
        assert second_gid_synapse_ids[3] == ('NodeB__NodeA__chemical', 3)



"""Graph representation of Cells and Synapses."""

from matplotlib.cm import ScalarMappable
from matplotlib.colors import Normalize
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np


from bluecellulab.cell.cell_dict import CellDict


def build_graph(cells: CellDict) -> nx.DiGraph:
    G = nx.DiGraph()

    # Add nodes (cells) to the graph
    for cell_id, cell in cells.items():
        G.add_node(cell_id)

    # Extract and add edges (connections) to the graph from each cell
    for cell_id, cell in cells.items():
        for connection in cell.connections.values():
            # Check if pre_cell exists for the connection
            if connection.pre_cell is None:
                continue

            # Source is the pre_cell from the connection
            source_cell_id = connection.pre_cell.cell_id

            # Target is the post-synapse cell from the connection
            target_cell_id = connection.post_synapse.post_cell_id

            # Check if both source and target cells are within the current cell collection
            if source_cell_id in cells and target_cell_id in cells:
                G.add_edge(source_cell_id, target_cell_id, weight=connection.weight)

    return G


def plot_graph(G: nx.Graph, node_size: float = 400, edge_width: float = 0.4, node_distance: float = 1.6):
    # Extract unique populations from the graph nodes
    populations = list(set([cell_id.population_name for cell_id in G.nodes()]))

    # Create a color map for each population
    color_map = plt.cm.tab20(np.linspace(0, 1, len(populations)))  # type: ignore[attr-defined]
    population_color = dict(zip(populations, color_map))

    # Create node colors based on their population
    node_colors = [population_color[node.population_name] for node in G.nodes()]

    # Extract weights for edge color mapping
    edge_weights = [d['weight'] for _, _, d in G.edges(data=True)]
    edge_colors = plt.cm.Greens(np.interp(edge_weights, (min(edge_weights), max(edge_weights)), (0.3, 1)))  # type: ignore[attr-defined]

    # Create positions using spring layout for the entire graph
    pos = nx.spring_layout(G, k=node_distance)

    # Create labels only for the node ID
    labels = {node: node.id for node in G.nodes()}

    # Create a figure and axis for the drawing
    fig, ax = plt.subplots(figsize=(6, 5))

    # Draw the graph
    nx.draw(G, pos, with_labels=True, labels=labels, node_color=node_colors,
            edge_color=edge_colors, width=edge_width, node_size=node_size, ax=ax, connectionstyle='arc3, rad = 0.1')

    # Draw directed edges
    nx.draw_networkx_edges(G, pos, edge_color=edge_colors, width=edge_width, ax=ax, arrowstyle='-|>', arrowsize=20, connectionstyle='arc3, rad = 0.1')

    # Create a legend
    for population, color in population_color.items():
        plt.plot([0], [0], color=color, label=population)
    plt.legend(loc="upper left", bbox_to_anchor=(-0.1, 1.05))  # Adjust these values as needed

    # Add a colorbar for edge weights
    sm = ScalarMappable(cmap="Greens", norm=Normalize(vmin=min(edge_weights), vmax=max(edge_weights)))
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax, orientation="vertical", fraction=0.03, pad=0.04)
    cbar.set_label('Synaptic Strength')

    # Add text at the bottom of the figure
    plt.figtext(0.5, 0.01, "Network of simulated cells", ha="center", fontsize=10, va="bottom")

    plt.show()



"""Unit tests for the synapse_factory module."""


import json
from pathlib import Path
import pandas as pd
import pytest
from bluecellulab.cell.core import Cell

from bluecellulab.circuit.circuit_access import EmodelProperties
from bluecellulab.circuit.config.sections import Conditions
from bluecellulab.circuit.synapse_properties import SynapseProperties, SynapseProperty, synapse_property_decoder
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.synapse import SynapseFactory
from bluecellulab.synapse.synapse_factory import SynapseType
from bluecellulab.synapse.synapse_types import GluSynapse


parent_dir = Path(__file__).resolve().parent.parent


@pytest.mark.v6
class TestSynapseFactory:

    def setup_method(self):
        emodel_properties = EmodelProperties(
            threshold_current=1.1433533430099487,
            holding_current=1.4146618843078613,
            AIS_scaler=1.4561502933502197,
            soma_scaler=1.0)
        self.cell = Cell(
            parent_dir / "examples/circuit_sonata_quick_scx/components/hoc/cADpyr_L2TPC.hoc",
            parent_dir / "examples/circuit_sonata_quick_scx/components/morphologies/asc/rr110330_C3_idA.asc",
            template_format="v6",
            emodel_properties=emodel_properties,
        )
        with open("tests/test_synapse/test-synapse-series.json") as f:
            self.syn_description = pd.Series(json.load(f, object_hook=synapse_property_decoder))

    def test_create_synapse(self):
        syn_id = ("a", 0)
        condition_parameters = Conditions.init_empty()
        popids = (0, 0)
        extracellular_calcium = 2.0
        connection_modifiers = {
            'Weight': 1.15143,
            'SpontMinis': 0.0,
            'SynapseConfigure': ['%s.Use = 1 %s.Use_GB = 1 %s.Use_p = 1 %s.gmax0_AMPA = gmax_p_AMPA %s.rho_GB = 1 %s.rho0_GB = 1 %s.gmax_AMPA = %s.gmax_p_AMPA']
        }

        synapse = SynapseFactory.create_synapse(
            self.cell, syn_id, self.syn_description, condition_parameters, popids, extracellular_calcium, connection_modifiers
        )
        assert isinstance(synapse, GluSynapse)
        assert synapse.weight == connection_modifiers["Weight"]

    def test_determine_synapse_location(self):
        res = SynapseFactory.determine_synapse_location(self.syn_description, self.cell)
        assert res.location == 0.9999999
        # set afferent_section_pos
        self.syn_description["afferent_section_pos"] = 1.2
        res = SynapseFactory.determine_synapse_location(self.syn_description, self.cell)
        assert res.location == 1.2
        assert res.section.L == pytest.approx(9.530376893488256)

    def test_synlocation_to_segx(self):
        ipt = 13
        isec = 169
        syn_offset = 1.2331762313842773
        section = self.cell.get_psection(isec).hsection
        res = SynapseFactory.synlocation_to_segx(section, ipt, syn_offset)
        assert res == pytest.approx(0.9999999)
        res = SynapseFactory.synlocation_to_segx(section, ipt, syn_offset=-1.0)
        assert res == pytest.approx(0.9999999)


def test_determine_synapse_type():
    # Mocking a pd.Series for syn_description
    syn_description = pd.Series({
        SynapseProperty.TYPE: 50,
        SynapseProperties.plasticity[0]: None,
        SynapseProperties.plasticity[1]: float('nan')
    })

    # Test when synapse is inhibitory
    assert SynapseFactory.determine_synapse_type(syn_description) == SynapseType.GABAAB

    # Test when synapse is excitatory with no plasticity
    syn_description[SynapseProperty.TYPE] = 150
    for prop in SynapseProperties.plasticity:
        syn_description[prop] = float('nan')  # All properties are NaN
    assert SynapseFactory.determine_synapse_type(syn_description) == SynapseType.AMPANMDA

    # Test when synapse is excitatory with all plasticity properties filled
    for prop in SynapseProperties.plasticity:
        syn_description[prop] = 1.0  # Assign a valid value to each plasticity property
    assert SynapseFactory.determine_synapse_type(syn_description) == SynapseType.GLUSYNAPSE

    # Test when synapse is excitatory with one plasticity property as NaN (should raise an exception)
    syn_description[SynapseProperties.plasticity[1]] = float('nan')
    with pytest.raises(BluecellulabError, match="SynapseFactory: Cannot determine synapse type"):
        SynapseFactory.determine_synapse_type(syn_description)



"""Unit tests for the psection module."""

import pytest
from pathlib import Path
from bluecellulab import Cell, EmodelProperties
from bluecellulab.psection import init_psections
from bluecellulab.cell.template import public_hoc_cell


parent_dir = Path(__file__).resolve().parent


class TestPSection:
    """Test class for testing Cell object functionalities with v6 template."""

    @pytest.fixture(autouse=True)
    def setup_method(self):
        """Setup."""
        emodel_properties = EmodelProperties(
            threshold_current=1.1433533430099487,
            holding_current=1.4146618843078613,
            AIS_scaler=1.4561502933502197,
            soma_scaler=1.0
        )
        self.cell = Cell(
            f"{parent_dir}/examples/circuit_sonata_quick_scx/components/hoc/cADpyr_L2TPC.hoc",
            f"{parent_dir}/examples/circuit_sonata_quick_scx/components/morphologies/asc/rr110330_C3_idA.asc",
            template_format="v6",
            emodel_properties=emodel_properties
        )
        self.psections, self.secname_to_psection = init_psections(public_hoc_cell(self.cell.cell))

    def test_is_leaf(self):
        """Test the is_leaf property for leaf and non-leaf sections."""
        leaf_section = self.psections[245]
        non_leaf_section = self.psections[1]
        assert leaf_section.is_leaf
        assert not non_leaf_section.is_leaf

    def test_hparent_and_hchildren(self):
        """Test hparent and hchildren properties for a section with a parent and children."""
        root = self.psections[0]
        assert root.hparent is None
        a_section = self.psections[145]
        assert a_section.hparent is not None
        assert len(a_section.hchildren) > 0
        leaf_section = self.psections[245]
        assert len(leaf_section.hchildren) == 0

    def test_add_pchild(self):
        """Test adding a child section."""
        parent_section = self.psections[245]  # leaf
        assert len(parent_section.pchildren) == 0
        child_section = self.psections[244]  # leaf
        parent_section.add_pchild(child_section)
        assert child_section in parent_section.pchildren
        assert child_section.pparent == parent_section

    def test_all_descendants(self):
        """Test retrieving all descendants of a section."""
        root_section = self.psections[0]
        all_descendants = root_section.all_descendants()
        # Excluding the root and myelin sections: -1 + -1 = -2
        assert len(all_descendants) == len(self.psections) - 2

    def test_tree_width(self):
        """Test calculation of tree width."""
        tree_width = self.psections[0].tree_width()
        assert tree_width > 0

    def test_tree_height(self):
        """Test calculation of tree height."""
        tree_height = self.psections[0].tree_height()
        assert tree_height > 0

    def test_init_psections(self):
        """Test if psections and secname_to_psection are properly initialized."""
        assert len(self.psections) > 0
        assert len(self.secname_to_psection) > 0
        assert len(self.psections) == len(self.secname_to_psection)



import json

import numpy as np
from bluecellulab.utils import CaptureOutput, IsolatedProcess, NumpyEncoder, run_once


# Decorated function for testing
@run_once
def increment_counter(counter):
    counter[0] += 1
    return "Executed"


def test_run_once_execution():
    """Test that the decorated function runs only once."""
    counter = [0]  # Using a list for mutability

    assert increment_counter(counter) == "Executed"
    increment_counter(counter)
    assert counter[0] == 1

    # Called 3 times but increased once
    increment_counter(counter)
    increment_counter(counter)
    increment_counter(counter)

    assert counter[0] == 1

    assert increment_counter(counter) is None


def test_capture_output():
    with CaptureOutput() as output:
        print("Hello, World!")
        print("This is a test.")

    assert len(output) == 2
    assert "Hello, World!" in output
    assert "This is a test." in output


def test_no_output():
    with CaptureOutput() as output:
        pass  # No output

    assert len(output) == 0


def test_numpy_encoder():
    """Utils: Test NumpyEncoder"""
    assert json.dumps(np.int32(1), cls=NumpyEncoder) == "1"
    assert json.dumps(np.float32(1.2), cls=NumpyEncoder)[0:3] == "1.2"
    assert json.dumps(np.array([1, 2, 3]), cls=NumpyEncoder) == "[1, 2, 3]"
    assert json.dumps(np.array([1.2, 2.3, 3.4]), cls=NumpyEncoder) == "[1.2, 2.3, 3.4]"
    assert (
        json.dumps(np.array([True, False, True]), cls=NumpyEncoder)
        == "[true, false, true]"
    )


def test_isolated_process():
    """Test to ensure isolated process keeps its properties."""
    runner = IsolatedProcess()
    assert runner._processes == 1
    assert runner._maxtasksperchild == 1



"""Unit tests for circuit::format.py."""
from pathlib import Path
from bluecellulab.circuit.format import CircuitFormat, determine_circuit_format, is_valid_json_file
from bluecellulab.circuit.config import SonataSimulationConfig

script_dir = Path(__file__).resolve().parent.parent

sonata_conf = (
    script_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "condition_parameters"
    / "simulation_config.json"
)
blueconfig = (
    script_dir
    / "examples"
    / "legacy_circuitconfig"
    / "CircuitConfig"
)


def test_determine_circuit_format():
    """Test the determine_circuit_format function."""
    assert determine_circuit_format(SonataSimulationConfig(sonata_conf)) == CircuitFormat.SONATA
    assert determine_circuit_format(sonata_conf) == CircuitFormat.SONATA
    assert determine_circuit_format(blueconfig) == CircuitFormat.BLUECONFIG


def test_is_valid_json_file():
    """Test the is_valid_json_file function."""
    assert not is_valid_json_file(blueconfig)
    assert not is_valid_json_file({1: 5, 2: 6})
    assert is_valid_json_file(sonata_conf)



"""Unit tests for circuit/config/sections.py."""

import pytest
from bluecellulab.circuit.config.sections import string_to_bool


def test_string_to_bool():
    """Unit test for the string_to_bool function."""
    assert string_to_bool("True")
    assert string_to_bool("true")
    assert string_to_bool("TRUE")
    assert string_to_bool("1")
    assert string_to_bool("False") is False
    assert string_to_bool("false") is False
    assert string_to_bool("FALSE") is False
    assert string_to_bool("0") is False
    with pytest.raises(ValueError):
        string_to_bool("invalid")



"""Unit tests for the simulation_access module."""

from pathlib import Path

import numpy as np
import pytest

from bluecellulab.circuit import CellId, SonataSimulationAccess
from bluecellulab.circuit.simulation_access import _sample_array, get_synapse_replay_spikes


parent_dir = Path(__file__).resolve().parent.parent

hipp_simulation_with_projections = (
    parent_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "projections"
    / "simulation_config.json"
)


def test_sample_array():
    """Unit test for _sample_array."""
    arr = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    t_step = 0.1
    sim_t_step = 0.1
    assert np.array_equal(_sample_array(arr, t_step / sim_t_step), arr)
    t_step = 0.2
    sim_t_step = 0.1
    assert np.array_equal(_sample_array(arr, t_step / sim_t_step), [1, 3, 5, 7, 9])
    t_step = 0.3
    sim_t_step = 0.1
    assert np.array_equal(_sample_array(arr, t_step / sim_t_step), [1, 4, 7, 10])
    t_step = 0.4
    sim_t_step = 0.1
    assert np.array_equal(_sample_array(arr, t_step / sim_t_step), [1, 5, 9])
    t_step = 0.1
    sim_t_step = 0.2
    with pytest.raises(ValueError):
        _sample_array(arr, t_step / sim_t_step)


class TestSonataSimulationAccess:

    def setup_method(self):
        self.simulation_access = SonataSimulationAccess(str(hipp_simulation_with_projections))

    def test_init_file_not_found(self):
        """Test BluepySimulationAccess init edge cases."""
        with pytest.raises(FileNotFoundError):
            SonataSimulationAccess(parent_dir / "examples" / "non_existing_file")

    def test_get_soma_voltage(self):
        """Test SonataCircuitAccess.get_soma_voltage."""
        cell_id = CellId("hippocampus_neurons", 1)
        t_start, t_end, t_step, sim_t_step = 0, 100, 0.1, 0.025
        soma_voltage = self.simulation_access.get_soma_voltage(cell_id, t_start, t_end, t_step)
        t_step_ratio = round(t_step / sim_t_step)
        assert soma_voltage.shape == (int((t_end - t_start) / t_step / t_step_ratio),)
        assert soma_voltage[0] == -65.0
        assert np.mean(soma_voltage) == pytest.approx(-54.447914)
        assert np.max(soma_voltage) == pytest.approx(41.53086)
        assert np.min(soma_voltage) == pytest.approx(-77.127716)

    def test_get_soma_time_trace(self):
        """Test SonataCircuitAccess.get_soma_time_trace."""
        t_step = 0.1
        soma_time_trace = self.simulation_access.get_soma_time_trace(t_step)
        assert soma_time_trace.shape == (250,)
        assert soma_time_trace[0] == 0.0
        assert soma_time_trace[1] == 0.4
        assert soma_time_trace[-1] == pytest.approx(99.6)

    def test_get_spikes(self):
        """Test SonataCircuitAccess.get_spikes."""
        spikes = self.simulation_access.get_spikes()
        assert len(spikes.keys()) == 10
        cell_id = CellId("hippocampus_neurons", 2)
        assert len(spikes[cell_id]) == 4
        np.testing.assert_almost_equal(
            spikes[cell_id], np.array([10.1, 22.65, 35.15, 47.675]))


def test_get_synapse_replay_spikes():
    """.Test get_synapse_replay_spikes."""
    res = get_synapse_replay_spikes(
        parent_dir / "data" / "synapse_replay_file" / "spikes.dat"
    )
    assert set(res.keys()) == {5382}
    assert res[5382].tolist() == [1500.0, 2000.0, 2500.0]



"""Unit tests for graph.py"""

import networkx as nx
from pathlib import Path

from matplotlib import pyplot as plt

from bluecellulab import SSim
from bluecellulab.graph import build_graph, plot_graph
from bluecellulab.circuit import CellId

script_dir = Path(__file__).parent


class TestGraph():
    """Test the graph.py module."""
    def setup_method(self):
        """Set up the test environment."""
        circuit_path = (
            script_dir
            / "examples/sim_quick_scx_sonata_multicircuit/simulation_config_hypamp.json"
        )
        self.sim = SSim(circuit_path)
        dstut_cells = [('NodeA', 0), ('NodeA', 1), ('NodeB', 0), ('NodeB', 1)]

        self.sim.instantiate_gids(dstut_cells, add_stimuli=True, add_synapses=True)
        t_stop = 50.0
        self.sim.run(t_stop)

        # Call the build_graph function
        self.graph = build_graph(self.sim.cells)

        self.expected_edges = [
            (CellId(population_name='NodeA', id=0), CellId(population_name='NodeB', id=0)),
            (CellId(population_name='NodeA', id=0), CellId(population_name='NodeA', id=1)),
            (CellId(population_name='NodeA', id=1), CellId(population_name='NodeA', id=0)),
            (CellId(population_name='NodeA', id=1), CellId(population_name='NodeB', id=0)),
            (CellId(population_name='NodeA', id=1), CellId(population_name='NodeB', id=1)),
            (CellId(population_name='NodeB', id=0), CellId(population_name='NodeB', id=1)),
            (CellId(population_name='NodeB', id=0), CellId(population_name='NodeA', id=0)),
            (CellId(population_name='NodeB', id=1), CellId(population_name='NodeA', id=0)),
            (CellId(population_name='NodeB', id=1), CellId(population_name='NodeB', id=0)),
        ]

    def test_graph_type(self):
        """Test if the graph is an instance of nx.DiGraph."""
        assert isinstance(self.graph, nx.DiGraph)

    def test_number_of_edges(self):
        """Test if the graph has the correct number of edges."""
        expected_number_of_edges = 9
        assert len(self.graph.edges) == expected_number_of_edges

    def test_existence_of_specific_edges(self):
        """Test for the existence of specific edges."""
        for edge in self.expected_edges:
            assert self.graph.has_edge(*edge), f"Edge {edge} is missing in the graph."

    def test_no_unexpected_edges(self):
        """Test if the graph does not have unexpected edges."""
        all_edges = set(self.graph.edges)
        assert all_edges.issubset(set(self.expected_edges)), "There are unexpected edges in the graph."

    def test_correct_edge_attributes(self):
        """Test if the edges have the correct weight."""
        edge = (CellId(population_name='NodeA', id=0), CellId(population_name='NodeB', id=0))
        expected_weight = 0.37
        actual_weight = self.graph.edges[edge]['weight']
        assert abs(expected_weight - actual_weight) <= 0.01, f"Edge {edge} has incorrect weight: expected {expected_weight}, found {actual_weight}."

    def test_plot_graph(self):
        """Test the plot_graph function"""
        plot_graph(self.graph)
        assert plt.fignum_exists(1), "No figure is created."



"""Unit tests for SSim."""
import numpy as np

import bluecellulab


def test_merge_pre_spike_trains():
    """SSim: Testing merge_pre_spike_trains"""

    train1 = {1: [5, 100], 2: [5, 8, 120]}
    train2 = {2: [7], 3: [8]}
    train3 = {1: [5, 100]}

    trains_merged = {1: [5, 5, 100, 100], 2: [5, 7, 8, 120], 3: [8]}

    np.testing.assert_equal(
        {},
        bluecellulab.ssim.SSim.merge_pre_spike_trains(None))
    np.testing.assert_equal(
        train1,
        bluecellulab.ssim.SSim.merge_pre_spike_trains(train1))
    np.testing.assert_equal(
        train1,
        bluecellulab.ssim.SSim.merge_pre_spike_trains(
            None,
            train1))
    np.testing.assert_equal(
        trains_merged,
        bluecellulab.ssim.SSim.merge_pre_spike_trains(
            train1,
            None,
            train2,
            train3))



"""Unit tests for synapse_properties module."""
from bluecellulab.circuit import SynapseProperty
from bluecellulab.circuit.synapse_properties import (
    properties_from_snap,
    properties_to_snap,
    synapse_property_decoder,
    synapse_property_encoder,
)


def test_synapse_property():
    """Test SynapseProperty."""
    pre_gid = SynapseProperty.PRE_GID
    assert SynapseProperty.from_snap("@source_node") == pre_gid
    assert SynapseProperty.to_snap(pre_gid) == "@source_node"


def test_properties_from_snap():
    """Test properties_from_snap."""
    snap_props = ["@source_node", "afferent_section_id", "rho0_GB"]
    props = properties_from_snap(snap_props)
    assert props == [
        SynapseProperty.PRE_GID,
        SynapseProperty.POST_SECTION_ID,
        "rho0_GB",
    ]


def test_properties_to_snap():
    """Test properties_to_snap."""
    props = [SynapseProperty.PRE_GID, SynapseProperty.POST_SECTION_ID, "rho0_GB"]
    snap_props = properties_to_snap(props)
    assert snap_props == ["@source_node", "afferent_section_id", "rho0_GB"]


def test_synapse_property_key_encoder():
    """Unit test for synapse_property_key_encoder."""

    # Sample input with a mix of SynapseProperty enum and string keys
    input_dict = {
        SynapseProperty.PRE_GID: 1234,
        SynapseProperty.POST_SECTION_ID: 5678,
        "sample_string_key": "sample_value"
    }

    # Expected output after encoding
    expected_output = {
        "PRE_GID": 1234,
        "POST_SECTION_ID": 5678,
        "sample_string_key": "sample_value"
    }

    assert synapse_property_encoder(input_dict) == expected_output


def test_synapse_property_decoder():
    """Unit test for synapse_property_decoder."""

    # Sample input with a mix of SynapseProperty enum (in string format) and string keys
    input_dict = {
        "PRE_GID": 1234,
        "POST_SECTION_ID": 5678,
        "sample_string_key": "sample_value"
    }

    # Expected output after decoding
    expected_output = {
        SynapseProperty.PRE_GID: 1234,
        SynapseProperty.POST_SECTION_ID: 5678,
        "sample_string_key": "sample_value"
    }

    assert synapse_property_decoder(input_dict) == expected_output



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represents a python version of NEURON Section (for drawing)."""
from __future__ import annotations
import neuron

import bluecellulab
from bluecellulab.cell.serialized_sections import SerializedSections
from bluecellulab.psegment import PSegment
from bluecellulab.type_aliases import HocObjectType, NeuronSection


def init_psections(
    hoc_cell: HocObjectType,
) -> tuple[dict[int, PSection], dict[str, PSection]]:
    """Initialize the psections list.

    This list contains the Python representation of the psections of
    this morphology.
    """
    psections: dict[int, PSection] = {}
    secname_to_psection: dict[str, PSection] = {}
    for sec in hoc_cell.all:
        secname = neuron.h.secname(sec=sec)
        secname_to_psection[secname] = PSection(sec)

    serial_sections = SerializedSections(hoc_cell)
    for isec, val in serial_sections.isec2sec.items():
        hsection: NeuronSection = val.sec
        if hsection:
            secname = neuron.h.secname(sec=hsection)
            psections[isec] = secname_to_psection[secname]
            psections[isec].isec = isec

    # Set the parents and children of all the psections
    for psec in psections.values():
        hparent = psec.hparent
        if hparent:
            parentname = hparent.name()
            psec.pparent = secname_to_psection[parentname]
        else:
            psec.pparent = None

        for hchild in psec.hchildren:
            childname = hchild.name()
            if "myelin" in childname:
                continue
            pchild = secname_to_psection[childname]
            psec.add_pchild(pchild)

    return psections, secname_to_psection


class PSection:
    """Class that represents a cell section."""

    def __init__(self, hsection: NeuronSection, isec: int | None = None):
        self.L: float = hsection.L
        self.diam: float = hsection.diam
        self.hsection = hsection
        self.name: str = neuron.h.secname(sec=hsection)
        self.href: NeuronSection = neuron.h.SectionRef(sec=self.hsection)
        self.pparent: PSection | None = None
        self.pchildren: list[PSection] = []
        self.isec = isec

        if 'apic' in self.name:
            self.section_type = 'apical'
        elif 'dend' in self.name:
            self.section_type = 'basal'
        elif 'soma' in self.name:
            self.section_type = 'somatic'
        elif 'axon' in self.name:
            self.section_type = 'axonal'
        elif 'myelin' in self.name:
            self.section_type = 'myelin'
        else:
            raise Exception(
                "PSection: Section of unknown type: %s" %
                self.name)

        self.psegments: list[PSegment] = []
        self.maxsegdiam = 0.0
        for hsegment in hsection:
            # psegment = bluecellulab.PSegment(hsection(hsegment.x), self)
            psegment = bluecellulab.PSegment(hsegment, self)
            self.psegments.append(psegment)
            self.maxsegdiam = max(self.maxsegdiam, psegment.diam)

        self.xSpacing = 1.0
        self.ySpacing = 5.0

    @property
    def is_leaf(self) -> bool:
        """Return true if section is a leaf in the morphological structure."""
        return not self.hchildren

    @property
    def hparent(self) -> NeuronSection | None:
        """Return the hoc section of the parent."""
        if self.href.has_parent():
            return self.href.parent
        else:
            return None

    @property
    def hchildren(self) -> list[NeuronSection]:
        """Return a list with the hoc sections of the children."""
        return [self.href.child[index] for index in
                range(0, int(self.href.nchild()))]

    def add_pchild(self, pchild: PSection) -> None:
        """Add a python represent of a child section."""
        self.pchildren.append(pchild)
        pchild.pparent = self

    def getSectionVarBounds(self, variable):
        """Get bounds a variable in a section."""
        varmin = None
        varmax = None
        for psegment in self.psegments:
            value = psegment.get_variable_value(variable)
            if value:
                varmin = min(value, varmin) if varmin else value
                varmax = max(value, varmax) if varmax else value
        return [varmin, varmax]

    def getTreeVarBounds(self, variable):
        """Get the bounds of a variable in a dendritic subtree."""
        varbounds = self.getSectionVarBounds(variable)
        for child in self.pchildren:
            child_varbounds = child.getTreeVarBounds(variable)
            if child_varbounds[0] and child_varbounds[1]:
                varbounds[0] = min(
                    varbounds[0],
                    child_varbounds[0]) if varbounds[0] else child_varbounds[0]
                varbounds[1] = max(
                    varbounds[1],
                    child_varbounds[1]) if varbounds[1] else child_varbounds[1]

        return varbounds

    def all_descendants(self) -> list[PSection]:
        """Return all the psection that are descendants of this psection."""
        pdescendants = list(self.pchildren)
        for child in self.pchildren:
            pdescendants += child.all_descendants()
        return pdescendants

    def tree_width(self) -> float:
        """Width of a dendritic tree."""
        if self.is_leaf:
            width = self.maxsegdiam + self.xSpacing
        else:
            width = sum(child.tree_width() for child in self.pchildren)
        return max(self.diam + self.xSpacing, width)

    def tree_height(self) -> float:
        """Height of dendritic tree."""
        return self.L + self.ySpacing + \
            (max([child.tree_height() for child in self.pchildren])
             if self.pchildren else 0)



"""Unit tests for the injector module."""

import math
from pathlib import Path

import neuron
import numpy as np
from pydantic import ValidationError
import pytest
from pytest import approx, raises

import bluecellulab

from bluecellulab.circuit import SonataCircuitAccess
from bluecellulab.circuit.node_id import CellId
from bluecellulab.stimulus.circuit_stimulus_definitions import (
    Pulse,
    Noise,
    Hyperpolarizing,
    RelativeLinear,
    OrnsteinUhlenbeck,
    RelativeOrnsteinUhlenbeck,
    ShotNoise,
    RelativeShotNoise,
    ClampMode,
)
from bluecellulab import RNGSettings
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.cell.stimuli_generator import gen_shotnoise_signal
from bluecellulab.cell import SonataProxy

script_dir = Path(__file__).resolve().parent.parent


@pytest.mark.v5
class TestInjector:
    """Test the InjectableMixin."""

    @classmethod
    def setup_method(cls):
        cls.cell = bluecellulab.Cell(f"{str(script_dir)}/examples/cell_example1/test_cell.hoc",
                                     f"{str(script_dir)}/examples/cell_example1/test_cell.asc")

        cls.sim = bluecellulab.Simulation()
        cls.sim.add_cell(cls.cell)

    def test_inject_pulse(self):
        """Test the pulse train injection."""
        stimulus = Pulse(
            target="single-cell",
            delay=2,
            duration=20,
            amp_start=4,
            frequency=5,
            width=2,
        )
        tstim = self.cell.add_pulse(stimulus)
        assert tstim.stim.to_python() == [0.0, 4.0, 4.0, 0.0, 0.0]
        assert tstim.tvec.to_python() == [2.0, 2.0, 4.0, 4.0, 22.0]

        with raises(ValidationError):
            unsupported_stimulus = Pulse(
                target="single-cell",
                delay=2,
                duration=20,
                amp_start=4,
                frequency=5,
                width=2,
                offset=1,
            )
            self.cell.add_pulse(unsupported_stimulus)

    def test_inject_step(self):
        """Test the step current injection."""
        tstim = self.cell.add_step(start_time=2.0, stop_time=6.0, level=1.0)
        assert tstim.stim.to_python() == [0.0, 1.0, 1.0, 0.0, 0.0]
        assert tstim.tvec.to_python() == [2.0, 2.0, 6.0, 6.0, 6.0]

    def test_inject_ramp(self):
        """Test the ramp injection."""
        tstim = self.cell.add_ramp(start_time=2.0, stop_time=6.0, start_level=0.5, stop_level=1)
        assert tstim.stim.to_python() == [0.0, 0.0, 0.5, 1.0, 0.0, 0.0]
        assert tstim.tvec.to_python() == [0.0, 2.0, 2.0, 6.0, 6.0, 6.0]

    def test_voltage_clamp(self):
        """Test adding voltage clamp."""
        amp = 1.5
        stop_time = 10
        rs = 1
        n_recording = "test_volt_clamp"
        seclamp_obj = self.cell.add_voltage_clamp(
            stop_time=stop_time, level=amp, rs=rs,
            current_record_name=n_recording
        )
        assert seclamp_obj.amp1 == amp
        assert seclamp_obj.dur1 == stop_time
        assert seclamp_obj.rs == rs

        self.sim.run(10, dt=1, cvode=False)
        current = self.cell.get_recording("test_volt_clamp")
        assert current == approx(np.array(
            [66.5, 5.39520998, -10.76796553, 20.6887735,
             17.8876999, 15.14995787, 13.47384441, 12.55945316,
             12.09052411, 11.8250991, 11.5502658]), abs=1e-3)

    def test_voltage_clamp_dt(self):
        """Test adding voltage clamp to a cell with a dt value."""
        amp = 1.5
        stop_time = 10
        rs = 1
        n_recording = "test_volt_clamp_dt"
        self.cell.record_dt = 2
        seclamp_obj = self.cell.add_voltage_clamp(
            stop_time=stop_time, level=amp, rs=rs,
            current_record_name=n_recording
        )
        assert seclamp_obj.amp1 == amp
        assert seclamp_obj.dur1 == stop_time
        assert seclamp_obj.rs == rs

        self.sim.run(10, dt=1, cvode=False)
        current = self.cell.get_recording("test_volt_clamp_dt")
        assert current == approx(np.array(
            [66.5, -10.76796553, 17.8876999, 13.47384441, 12.09052411]), abs=1e-3)

    def test_get_noise_step_rand(self):
        """Unit test for _get_noise_step_rand."""
        noisestim_count = 5
        for mode in ["Compatibility", "UpdatedMCell", "Random123"]:
            rng_obj = RNGSettings.get_instance()
            rng_obj.mode = mode
            rng = self.cell._get_noise_step_rand(noisestim_count)
        rng.Random123()
        rng.poisson(4)
        rng.MCellRan4()
        assert rng.uniform(1, 15) == 2.9755221367813647

    def test_add_noise_step(self):
        """Test adding a step current with noise on top."""
        rng_obj = RNGSettings.get_instance()
        rng_obj.mode = "Compatibility"
        tstim = self.cell.add_noise_step(
            section=self.cell.soma, segx=0.5, mean=2, variance=0.05, delay=2,
            duration=8, noisestim_count=5)
        assert tstim.stim.as_numpy() == approx(np.array(
            [0.0, 1.688244150056253,
             1.5206057728916358, 1.963561732554084, 1.5287573262092622,
             1.8136627185053698, 2.1230204494073135, 1.8715777361739463,
             1.7068988305615118, 1.7574514888132944, 2.055318487170783,
             1.8673307717912755, 1.932569903725156, 1.9394341839268754,
             1.8843667144133713, 1.8175816051992186, 1.927545675194812, 0.0]))
        assert tstim.tvec.to_python() == [
            2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0,
            5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0, 0.0]

    def test_add_noise_step_with_seed(self):
        """Test adding a step current with noise on top with a seed."""
        rng_obj = RNGSettings.get_instance()
        rng_obj.mode = "Compatibility"
        tstim = self.cell.add_noise_step(
            section=self.cell.soma, segx=0.5, mean=2, variance=0.05, delay=2,
            duration=8, seed=1)

        assert tstim.stim.as_numpy() == approx(np.array(
            [0., 1.84104848, 1.97759473, 2.22855241, 1.80930735,
             2.09799701, 2.10379869, 2.29691643, 2.26258353, 2.14120033,
             1.93326057, 1.94724241, 1.87856356, 2.4008308, 1.91991524,
             1.50814262, 1.83374623, 0.]))

        assert tstim.tvec.to_python() == [
            2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0,
            5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0, 0.0]

    def test_add_replay_noise(self):
        """Unit test for add_replay_noise."""
        rng_obj = RNGSettings.get_instance()
        rng_obj.mode = "Compatibility"
        stimulus = Noise(
            mean_percent=1, variance=0.1, delay=4, duration=10, target="single-cell"
        )
        tstim = self.cell.add_replay_noise(stimulus, noise_seed=1)
        assert tstim.stim.as_numpy() == approx(np.array(
            [0., -0.00780348, 0.00048122, 0.01570763, -0.00972932,
             0.00778641, 0.00813842, 0.0198555, 0.01777241, 0.0104077,
             -0.00220868, -0.00136035, -0.00552732, 0.02616032, -0.00301838,
             -0.02800195, -0.00824653, -0.00273605, 0.00022639, 0.009682,
             0.00787559, 0.]), abs=1e-5)
        assert tstim.tvec.to_python() == [
            4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0,
            7.5, 8.0, 8.5, 9.0, 9.5, 10.0, 10.5, 11.0, 11.5, 12.0, 12.5, 13.0,
            13.5, 14.0, 0.0]

    def test_add_replay_hypamp(self):
        """Unit test for add_replay_hypamp method."""
        stimulus = Hyperpolarizing(target="single-cell", delay=4, duration=20)
        tstim = self.cell.add_replay_hypamp(stimulus)
        hypamp = self.cell.hypamp
        assert tstim.stim.to_python() == [0.0, hypamp, hypamp, 0.0, 0.0]
        assert tstim.tvec.to_python() == [4.0, 4.0, 24.0, 24.0, 24.0]
        with pytest.raises(BluecellulabError):
            self.cell.hypamp = None
            self.cell.add_replay_hypamp(stimulus)

    def test_add_replay_relativelinear(self):
        """Unit test for add_replay_relativelinear."""
        stimulus = RelativeLinear(
            target="single-cell",
            delay=4, duration=20, percent_start=60)
        tstim = self.cell.add_replay_relativelinear(stimulus)
        assert tstim.stim.to_python() == [0.0, 0.1104372, 0.1104372, 0.0, 0.0]
        assert tstim.tvec.to_python() == [4.0, 4.0, 24.0, 24.0, 24.0]

    def test_get_ornstein_uhlenbeck_rand(self):
        """Unit test to check RNG generated for ornstein_uhlenbeck."""
        rng = self.cell._get_ornstein_uhlenbeck_rand(0, 144)
        assert rng.uniform(1, 15) == 12.477080945047298

        with raises(BluecellulabError):
            rng_obj = RNGSettings.get_instance()
            rng_obj.mode = "Compatibility"
            self.cell._get_ornstein_uhlenbeck_rand(0, 144)

    def test_get_shotnoise_step_rand(self):
        """Unit test to check RNG generated for shotnoise."""
        rng = self.cell._get_shotnoise_step_rand(0, 144)
        assert rng.uniform(1, 15) == 7.260484082563668

        with raises(BluecellulabError):
            rng_obj = RNGSettings.get_instance()
            rng_obj.mode = "Compatibility"
            self.cell._get_shotnoise_step_rand(0, 144)

    def test_add_replay_shotnoise(self):
        """Unit test for add_replay_shotnoise."""
        rng_obj = RNGSettings.get_instance()
        rng_obj.set_seeds(base_seed=549821, mode="Random123")
        rng_obj.stimulus_seed = 549821
        soma = self.cell.soma
        segx = 0.5
        stimulus = ShotNoise(
            target="single-cell", delay=0, duration=2,
            rise_time=0.4, decay_time=4, rate=2E3, amp_mean=40E-3, amp_var=16E-4,
            seed=3899663
        )
        time_vec, stim_vec = self.cell.add_replay_shotnoise(soma, segx, stimulus,
                                                            shotnoise_stim_count=3)
        assert list(time_vec) == approx([0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5,
                                         1.75, 2.0, 2.0])
        assert list(stim_vec) == approx([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0077349976,
                                         0.0114066037, 0.0432062144, 0.0])
        with raises(ValidationError):
            invalid_stim = ShotNoise(
                target="single-cell", delay=0, duration=2,
                rise_time=4.4, decay_time=4, rate=2E3, amp_mean=40E-3, amp_var=16E-4,
                seed=3899663
            )
            self.cell.add_replay_shotnoise(soma, segx, invalid_stim, shotnoise_stim_count=3)

    def test_add_ornstein_uhlenbeck(self):
        """Unit test for add_ornstein_uhlenbeck."""
        rng_obj = RNGSettings.get_instance()
        rng_obj.set_seeds(base_seed=549821, mode="Random123")
        rng_obj.stimulus_seed = 549821
        soma = self.cell.soma
        segx = 0.5
        stimulus = OrnsteinUhlenbeck(
            target="single-cell", delay=0, duration=2,
            tau=2.8, sigma=0.0042, mean=0.029, mode="current_clamp", dt=0.25, seed=1
        )
        time_vec, stim_vec = self.cell.add_ornstein_uhlenbeck(soma, segx, stimulus,
                                                              stim_count=1)

        assert list(time_vec) == approx([0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5,
                                         1.75, 2.0, 2.0])
        assert list(stim_vec) == approx([0.029, 0.030066357907250596, 0.0292377422585839,
                                         0.02740869851318549, 0.02615778315096319,
                                         0.026618755144424376, 0.026643574535378675,
                                         0.027504132283051937, 0.028251366180467998, 0.0])

        stimulus = OrnsteinUhlenbeck(
            target="single-cell", delay=0, duration=2,
            tau=2.8, sigma=0.0042, mean=0.029, mode="conductance", dt=0.25, seed=1
        )
        time_vec, stim_vec = self.cell.add_ornstein_uhlenbeck(soma, segx, stimulus,
                                                              stim_count=1)

        assert list(time_vec) == approx([0, 0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5,
                                         1.75, 2.0, 2.0])
        assert list(stim_vec) == approx([1000000000.0, 34.48275862068965, 33.259765053180814,
                                         34.20236730852261, 36.48476776520164, 38.22953933935253,
                                         37.56749684853171, 37.5325014544182, 36.35817300865007,
                                         35.39651829975447, 1000000000.0])

    def test_inject_current_clamp_signal(self):
        """Unit test for inject_current_clamp_signal."""
        tvec = neuron.h.Vector(np.arange(10))
        svec = neuron.h.Vector(np.random.normal(0, 0.1, 10))

        original_tvec = neuron.h.Vector().copy(tvec)
        original_svec = neuron.h.Vector().copy(svec)

        soma = self.cell.soma
        segx = 0.5
        self.cell.inject_current_clamp_signal(soma, segx, tvec, svec)

        assert len(self.cell.persistent) == 3
        assert "IClamp" in self.cell.persistent[0].hname()
        assert self.cell.persistent[0].dur == tvec[-1]  # check duration
        # tvec and svec are not modified
        assert original_tvec.eq(self.cell.persistent[1]) == 1.0
        assert original_svec.eq(self.cell.persistent[2]) == 1.0

    def test_inject_current_clamp_via_shotnoise_signal(self):
        """Unit test for inject_current_clamp_signal using a shotnoise_step."""
        soma = self.cell.soma
        segx = 0.5
        rng = self.cell._get_shotnoise_step_rand(shotnoise_stim_count=0, seed=None)
        tvec, svec = gen_shotnoise_signal(tau_D=4.0, tau_R=0.4, rate=2e3, amp_mean=40e-3,
                                          amp_var=16e-4, duration=2, dt=0.25, rng=rng)
        delay = 0
        tvec.add(delay)  # add delay
        tvec, svec = self.cell.inject_current_clamp_signal(soma, segx, tvec, svec)

        assert svec.as_numpy() == approx(np.array(
            [0., 0., 0., 0.00822223, 0.01212512,
             0.0137462, 0.034025, 0.04967694, 0.05614846, 0.]))
        assert tvec.to_python() == approx(
            [0.0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75, 2.0, 2.0])

    def test_inject_dynamic_clamp_signal(self):
        """Unit test for inject_dynamic_clamp_signal."""
        tvec = neuron.h.Vector(np.arange(10))
        svec = neuron.h.Vector(np.random.normal(0, 0.1, 10))

        original_tvec = neuron.h.Vector().copy(tvec)
        original_svec = neuron.h.Vector().copy(svec)

        soma = self.cell.soma
        segx = 0.5
        reversal = 1e-3
        self.cell.inject_dynamic_clamp_signal(soma, segx, tvec, svec, reversal)
        assert len(self.cell.persistent) == 3
        assert "SEClamp" in self.cell.persistent[0].hname()
        assert self.cell.persistent[0].amp1 == reversal
        assert self.cell.persistent[0].dur1 == tvec[-1]  # check duration
        # tvec and svec are modified
        assert original_tvec.eq(self.cell.persistent[1]) != 1.0
        assert original_svec.eq(self.cell.persistent[2]) != 1.0

    def test_add_replay_relative_shotnoise(self):
        """Unit test for add_replay_relative_shotnoise."""
        stimulus = RelativeShotNoise(
            target="single-cell", delay=0, duration=2,
            rise_time=0.4, decay_time=4, mean_percent=70, sd_percent=40, amp_cv=0.63,
            seed=12,
        )
        self.cell.threshold = 0.184062
        soma = self.cell.soma
        segx = 0.5
        tvec, svec = self.cell.add_replay_relative_shotnoise(soma, segx, stimulus)
        assert svec.to_python() == approx([0., 0., 0., 0., 0.0204470197, 0.0301526984,
                                          0.0341840080, 0.0352485557, 0.0347913472, 0.])
        assert tvec.to_python() == approx([0.0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75,
                                          2.0, 2.0])

        with raises(ValidationError):
            invalid_stim = RelativeShotNoise(
                target="single-cell", delay=0, duration=2,
                rise_time=4, decay_time=4, mean_percent=70, sd_percent=40, amp_cv=0.63,
                seed=12,
            )
            self.cell.add_replay_relative_shotnoise(soma, segx, invalid_stim)

        with raises(ZeroDivisionError):
            self.cell.threshold = 0.0
            self.cell.add_replay_relative_shotnoise(soma, segx, stimulus)

    def test_inject_current_waveform(self):
        """Test injecting any input current and time arrays."""
        start_time, stop_time, dt = 10.0, 20.0, 1.0
        amplitude, freq, mid_level = 12.0, 50.0, 0.0

        t_content = np.arange(start_time, stop_time, dt)
        i_content = [amplitude * math.sin(freq * (x - start_time) * (
            2 * math.pi)) + mid_level for x in t_content]

        current = self.cell.inject_current_waveform(t_content, i_content)
        assert current.as_numpy() == approx(np.array(
            [0.00000000e+00, 2.35726407e-14, 4.71452814e-14, -6.11403104e-13,
             9.42905627e-14, -1.92849988e-12, -1.22280621e-12, -3.24559665e-12,
             1.88581125e-13, -1.83420931e-12]), abs=1e-3)

    def test_add_sin_current(self):
        """Unit test for add_sin_current."""
        start_time, duration = 2, 10
        amp, freq = 12.0, 50.0
        tstim = self.cell.add_sin_current(amp, start_time, duration, freq)
        assert tstim.stim.as_numpy()[195:205] == approx(np.array(
            [11.99074843, 11.99407872, 11.99666916, 11.99851959, 11.99962989,
             12., 11.99962989, 11.99851959, 11.99666916, 11.99407872]))


class TestInjectorSonata:
    """Test the InjectableMixin on local Sonata circuit."""

    @classmethod
    def setup_method(cls):
        circuit_path = (
            script_dir
            / "examples"
            / "sonata_unit_test_sims"
            / "condition_parameters"
            / "simulation_config.json"
        )
        circuit_access = SonataCircuitAccess(circuit_path)
        cell_id = CellId("NodeA", 1)
        hoc = circuit_access.emodel_path(cell_id)
        template_format = circuit_access.get_template_format()
        morph = circuit_access.morph_filepath(cell_id)
        emodel_properties = circuit_access.get_emodel_properties(cell_id)
        rng_settings = RNGSettings.get_instance()
        rng_settings.set_seeds(sim_config=circuit_access.config)
        cls.cell = bluecellulab.Cell(template_path=hoc, morphology_path=morph,
                                     template_format=template_format, emodel_properties=emodel_properties)

        sonata_proxy = SonataProxy(cell_id, circuit_access)
        cls.cell.connect_to_circuit(sonata_proxy)

    @pytest.mark.v6
    def test_add_relative_ornstein_uhlenbeck_sd_error(self):
        """Unit test for add_relative_ornstein_uhlenbeck with 0 sd exception."""
        stimulus = RelativeOrnsteinUhlenbeck(
            target="single-cell", delay=0,
            duration=2, tau=2.8, mean_percent=3.078, sd_percent=0,
            mode=ClampMode.CURRENT, dt=0.25, seed=1
        )
        soma = self.cell.soma
        segx = 0.5
        with pytest.raises(BluecellulabError, match="standard deviation: 0.0, must be positive"):
            self.cell.add_relative_ornstein_uhlenbeck(soma, segx, stimulus)

    @pytest.mark.v6
    def test_add_relative_ornstein_uhlenbeck(self):
        """Unit test for adding relative ornstein_uhlenbeck."""
        stimulus = RelativeOrnsteinUhlenbeck(
            target="single-cell", delay=0,
            duration=2, tau=2.8, mean_percent=3.078, sd_percent=0.6156,
            mode=ClampMode.CURRENT, dt=0.25, seed=1
        )
        soma = self.cell.soma
        segx = 0.5
        time_vec, stim_vec = self.cell.add_relative_ornstein_uhlenbeck(
            soma, segx, stimulus)
        assert list(time_vec) == approx([0, 0.25, 0.5, 0.75, 1.0, 1.25, 1.5,
                                         1.75, 2.0, 2.0])
        assert list(stim_vec) == approx(
            [0.050476, 0.048279, 0.049130, 0.050349, 0.049961, 0.054486,
             0.061342, 0.051148, 0.047710, 0.0], abs=1e-6)

    @pytest.mark.v6
    def test_add_alpha_synapse(self):
        """Unit test for adding alpha synapse stimulus to cell."""
        onset = 2.0
        tau = 10.0
        gmax = 0.0003
        e = 0.0
        section = self.cell.cell.getCell().dend[15]
        syn = self.cell.add_alpha_synapse(onset, tau, gmax, e, section)

        current_vector = neuron.h.Vector()
        current_vector.record(syn._ref_i)
        sim = bluecellulab.Simulation()
        sim.add_cell(self.cell)
        sim.run(6.0, cvode=False, dt=1)
        current_vector = current_vector.to_python()
        # assert first 2 values are 0
        assert current_vector[0] == approx(0.0)
        assert current_vector[1] == approx(0.0)
        # assert last 3 values less than 0
        for j in range(-3, 0):
            assert current_vector[j] < 0.0



"""Utility functions used within BlueCellulab."""
from __future__ import annotations
import contextlib
import io
import json
from multiprocessing.pool import Pool

import numpy as np


def run_once(func):
    """A decorator to ensure a function is only called once."""
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return func(*args, **kwargs)
    wrapper.has_run = False
    return wrapper


class CaptureOutput(list):
    def __enter__(self):
        self._stringio = io.StringIO()
        self._redirect_stdout = contextlib.redirect_stdout(self._stringio)
        self._redirect_stdout.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._redirect_stdout.__exit__(exc_type, exc_val, exc_tb)
        self.extend(self._stringio.getvalue().splitlines())


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class IsolatedProcess(Pool):
    """Multiprocessing Pool that restricts a worker to run max 1 process.

    Use this when running isolated NEURON simulations. Running 2 NEURON
    simulations on a single process is to be avoided.
    """
    def __init__(self):
        super().__init__(processes=1, maxtasksperchild=1)



"""Unit tests for the circuit_access module."""

from pathlib import Path
from unittest.mock import patch

import pandas as pd
import pytest

from bluecellulab.circuit import CellId, SonataCircuitAccess
from bluecellulab.circuit.circuit_access import EmodelProperties, get_synapse_connection_parameters
from bluecellulab.circuit import SynapseProperty


parent_dir = Path(__file__).resolve().parent.parent

hipp_circuit_with_projections = (
    parent_dir
    / "examples"
    / "sonata_unit_test_sims"
    / "projections"
    / "simulation_config.json"
)


def test_get_synapse_connection_parameters():
    """Test that the synapse connection parameters are correctly created."""
    circuit_access = SonataCircuitAccess(hipp_circuit_with_projections)
    # both cells are Excitatory
    pre_cell_id = CellId("hippocampus_neurons", 1)
    post_cell_id = CellId("hippocampus_neurons", 2)
    connection_params = get_synapse_connection_parameters(
        circuit_access, pre_cell_id, post_cell_id)
    assert connection_params["add_synapse"] is True
    assert connection_params["Weight"] == 1.0
    assert connection_params["SpontMinis"] == 0.01
    syn_configure = connection_params["SynapseConfigure"]
    assert syn_configure[0] == "%s.NMDA_ratio = 1.22 %s.tau_r_NMDA = 3.9 %s.tau_d_NMDA = 148.5"
    assert syn_configure[1] == "%s.mg = 1.0"


def test_sonata_circuit_access_file_not_found():
    with pytest.raises(FileNotFoundError):
        SonataCircuitAccess("non_existing_file")


class TestSonataCircuitAccess:
    def setup_method(self):
        self.circuit_access = SonataCircuitAccess(hipp_circuit_with_projections)

    def test_available_cell_properties(self):
        assert self.circuit_access.available_cell_properties == {
            "x",
            "region",
            "rotation_angle_yaxis",
            "etype",
            "model_template",
            "synapse_class",
            "@dynamics:holding_current",
            "morph_class",
            "population",
            "morphology",
            "y",
            "layer",
            "mtype",
            "rotation_angle_xaxis",
            "z",
            "@dynamics:threshold_current",
            "model_type",
            "rotation_angle_zaxis",
        }

    def test_get_emodel_properties(self):
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.get_emodel_properties(cell_id)
        assert res == EmodelProperties(
            threshold_current=0.33203125, holding_current=-0.116351104071555
        )

    def test_get_template_format(self):
        res = self.circuit_access.get_template_format()
        assert res == "v6"

    def test_get_cell_properties(self):
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.get_cell_properties(cell_id, "layer")
        assert res.values == ["SP"]

    def test_extract_synapses(self):
        cell_id = CellId("hippocampus_neurons", 1)
        projections = None
        res = self.circuit_access.extract_synapses(cell_id, projections)
        assert res.shape == (1742, 15)
        assert all(res["source_popid"] == 2126)

        assert all(res["source_population_name"] == "hippocampus_projections")
        assert all(res["target_popid"] == 378)
        assert all(res[SynapseProperty.POST_SEGMENT_ID] != -1)
        assert SynapseProperty.U_HILL_COEFFICIENT not in res.columns
        assert SynapseProperty.CONDUCTANCE_RATIO not in res.columns

        # projection parameter
        projection = "hippocampus_projections"
        res = self.circuit_access.extract_synapses(cell_id, projection)
        assert res.shape == (1742, 15)
        list_of_single_projection = [projection]
        res = self.circuit_access.extract_synapses(cell_id, list_of_single_projection)
        assert res.shape == (1742, 15)
        empty_projection = []
        res = self.circuit_access.extract_synapses(cell_id, empty_projection)
        assert res.shape == (1742, 15)

    def test_target_contains_cell(self):
        target = "most_central_10_SP_PC"
        cell = CellId("hippocampus_neurons", 1)
        assert self.circuit_access.target_contains_cell(target, cell)
        cell2 = CellId("hippocampus_neurons", 100)
        assert not self.circuit_access.target_contains_cell(target, cell2)

    def test_get_target_cell_ids(self):
        target = "most_central_10_SP_PC"
        res = self.circuit_access.get_target_cell_ids(target)
        res_populations = {x.population_name for x in res}
        res_ids = {x.id for x in res}
        assert res_populations == {'hippocampus_neurons'}
        assert res_ids == {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}

    def test_get_cell_ids_of_targets(self):
        targets = ["most_central_10_SP_PC", "Mosaic"]
        res = self.circuit_access.get_cell_ids_of_targets(targets)
        res_ids = {x.id for x in res}
        assert res_ids == {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}

    def test_get_gids_of_mtypes(self):
        res = self.circuit_access.get_gids_of_mtypes(["SP_PC"])
        res_ids = {x.id for x in res}
        assert res_ids == {0, 1, 2, 3, 4, 5, 6, 7, 8, 9}

    def test_morph_filepath(self):
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.morph_filepath(cell_id)
        fname = "dend-mpg141216_A_idA_axon-mpg141017_a1-2_idC"\
            "_-_Scale_x1.000_y0.900_z1.000_-_Clone_12.swc"
        assert res.rsplit("/", 1)[-1] == fname

    def test_emodel_path(self):
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.emodel_path(cell_id)
        fname = "CA1_pyr_cACpyr_mpg141017_a1_2_idC_2019032814340.hoc"
        assert res.rsplit("/", 1)[-1] == fname

    def test_is_valid_group(self):
        group = "most_central_10_SP_PC"
        assert self.circuit_access.is_valid_group(group)
        group2 = "most_central_10_SP_PC2"
        assert not self.circuit_access.is_valid_group(group2)

    def test_fetch_cell_info(self):
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.fetch_cell_info(cell_id)
        assert res.shape == (18,)
        assert res.etype == "cACpyr"
        assert res.mtype == "SP_PC"
        assert res.rotation_angle_zaxis == pytest.approx(-3.141593)

    def test_fetch_mini_frequencies(self):
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.fetch_mini_frequencies(cell_id)
        assert res == (None, None)

    def test_node_properties_available(self):
        assert self.circuit_access.node_properties_available

    @patch("bluecellulab.circuit.SonataCircuitAccess.fetch_cell_info")
    def test_fetch_mini_frequencies_with_mock(self, mock_fetch_cell_info):
        mock_fetch_cell_info.return_value = pd.Series(
            {
                "exc-mini_frequency": 0.03,
                "inh-mini_frequency": 0.05,
            }
        )
        cell_id = CellId("hippocampus_neurons", 1)
        res = self.circuit_access.fetch_mini_frequencies(cell_id)
        assert res == (0.03, 0.05)

    def test_get_population_ids(self):
        source = "hippocampus_projections"
        target = "hippocampus_neurons"
        source_popid, target_popid = self.circuit_access.get_population_ids(source, target)
        assert source_popid == 2126
        assert target_popid == 378


def test_morph_filepath_asc():
    """Unit test for SonataCircuitAccess::morph_filepath for asc retrieval."""
    circuit_sonata_quick_scx_config = (
        parent_dir
        / "examples"
        / "sonata_unit_test_sims"
        / "condition_parameters"
        / "simulation_config.json"
    )

    circuit_access = SonataCircuitAccess(circuit_sonata_quick_scx_config)
    asc_morph = circuit_access.morph_filepath(CellId("NodeA", 1))
    assert asc_morph.endswith(".asc")


def test_get_emodel_properties_soma_scaler():
    """Test the retrieval of soma scaler value."""
    circuit_sonata_quick_scx_config = (
        parent_dir
        / "examples"
        / "sonata_unit_test_sims"
        / "condition_parameters"
        / "simulation_config.json"
    )

    circuit_access = SonataCircuitAccess(circuit_sonata_quick_scx_config)
    assert circuit_access.get_emodel_properties(CellId("NodeA", 0)).soma_scaler == 1.0
    assert circuit_access.get_emodel_properties(CellId("NodeA", 1)).soma_scaler == 1.002



# Copyright 2012-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cell class."""

from __future__ import annotations

import logging

from pathlib import Path
import queue
from typing import Optional
from typing_extensions import deprecated

import neuron
import numpy as np
import pandas as pd

import bluecellulab
from bluecellulab.psection import PSection, init_psections
from bluecellulab.cell.injector import InjectableMixin
from bluecellulab.cell.plotting import PlottableMixin
from bluecellulab.cell.section_distance import EuclideanSectionDistance
from bluecellulab.cell.sonata_proxy import SonataProxy
from bluecellulab.cell.template import NeuronTemplate, public_hoc_cell
from bluecellulab.circuit.config.sections import Conditions
from bluecellulab.circuit import EmodelProperties, SynapseProperty
from bluecellulab.circuit.node_id import CellId
from bluecellulab.circuit.simulation_access import get_synapse_replay_spikes
from bluecellulab.exceptions import BluecellulabError
from bluecellulab.importer import load_hoc_and_mod_files
from bluecellulab.neuron_interpreter import eval_neuron
from bluecellulab.rngsettings import RNGSettings
from bluecellulab.stimulus.circuit_stimulus_definitions import SynapseReplay
from bluecellulab.synapse import SynapseFactory, Synapse
from bluecellulab.synapse.synapse_types import SynapseID
from bluecellulab.type_aliases import HocObjectType, NeuronSection

logger = logging.getLogger(__name__)


class Cell(InjectableMixin, PlottableMixin):
    """Represents a Cell object."""

    last_id = 0

    @load_hoc_and_mod_files
    def __init__(self,
                 template_path: str | Path,
                 morphology_path: str | Path,
                 cell_id: Optional[CellId] = None,
                 record_dt: Optional[float] = None,
                 template_format: str = "v5",
                 emodel_properties: Optional[EmodelProperties] = None) -> None:
        """Initializes a Cell object.

        Args:
            template_path: Path to hoc template file.
            morphology_path: Path to morphology file.
            cell_id: ID of the cell, used in RNG seeds.
            record_dt: Timestep for the recordings.
            template_format: Cell template format such as 'v5' or 'v6_air_scaler'.
            emodel_properties: Template specific emodel properties.
        """
        super().__init__()
        if cell_id is None:
            cell_id = CellId("", Cell.last_id)
            Cell.last_id += 1
        self.cell_id = cell_id

        # Load the template
        neuron_template = NeuronTemplate(template_path, morphology_path, template_format, emodel_properties)
        self.template_id = neuron_template.template_name  # useful to map NEURON and python objects
        self.cell = neuron_template.get_cell(self.cell_id.id)
        if template_format == 'v6':
            if emodel_properties is None:
                raise BluecellulabError('EmodelProperties must be provided for v6 template')
            self.hypamp: float | None = emodel_properties.holding_current
            self.threshold: float | None = emodel_properties.threshold_current
        else:
            try:
                self.hypamp = self.cell.getHypAmp()
            except AttributeError:
                self.hypamp = None

            try:
                self.threshold = self.cell.getThreshold()
            except AttributeError:
                self.threshold = None
        self.soma = public_hoc_cell(self.cell).soma[0]
        # WARNING: this finitialize 'must' be here, otherwhise the
        # diameters of the loaded morph are wrong
        neuron.h.finitialize()

        self.recordings: dict[str, HocObjectType] = {}
        self.synapses: dict[SynapseID, Synapse] = {}
        self.connections: dict[SynapseID, bluecellulab.Connection] = {}

        self.ips: dict[SynapseID, HocObjectType] = {}
        self.syn_mini_netcons: dict[SynapseID, HocObjectType] = {}

        # Be careful when removing this,
        # time recording needs this push
        self.soma.push()
        self.hocname = neuron.h.secname(sec=self.soma).split(".")[0]
        self.record_dt = record_dt
        self.add_recordings(['self.soma(0.5)._ref_v', 'neuron.h._ref_t'],
                            dt=self.record_dt)

        self.delayed_weights = queue.PriorityQueue()  # type: ignore
        self.psections, self.secname_to_psection = init_psections(public_hoc_cell(self.cell))

        # Keep track of when a cell is made passive by make_passive()
        # Used to know when re_init_rng() can be executed
        self.is_made_passive = False

        neuron.h.pop_section()  # Undoing soma push
        self.sonata_proxy: Optional[SonataProxy] = None

        # Persistent objects, like clamps, that exist as long
        # as the object exists
        self.persistent: list[HocObjectType] = []

    @property
    def somatic(self) -> list[NeuronSection]:
        return list(public_hoc_cell(self.cell).somatic)

    @property
    def basal(self) -> list[NeuronSection]:
        return list(public_hoc_cell(self.cell).basal)

    @property
    def apical(self) -> list[NeuronSection]:
        return list(public_hoc_cell(self.cell).apical)

    @property
    def axonal(self) -> list[NeuronSection]:
        return list(public_hoc_cell(self.cell).axonal)

    @property
    def all(self) -> list[NeuronSection]:
        return list(public_hoc_cell(self.cell).all)

    def __repr__(self) -> str:
        base_info = f"Cell Object: {super().__repr__()}"
        hoc_info = f"NEURON ID: {self.template_id}"
        return f"{base_info}.\n{hoc_info}."

    def connect_to_circuit(self, sonata_proxy: SonataProxy) -> None:
        """Connect this cell to a circuit via sonata proxy."""
        self.sonata_proxy = sonata_proxy

    def re_init_rng(self, use_random123_stochkv: bool = False) -> None:
        """Reinitialize the random number generator for stochastic channels."""

        if not self.is_made_passive:
            if use_random123_stochkv:
                channel_id = 0
                for section in self.somatic:
                    for seg in section:
                        neuron.h.setdata_StochKv(seg.x, sec=section)
                        neuron.h.setRNG_StochKv(channel_id, self.cell_id.id)
                        channel_id += 1
                for section in self.basal:
                    for seg in section:
                        neuron.h.setdata_StochKv(seg.x, sec=section)
                        neuron.h.setRNG_StochKv(channel_id, self.cell_id.id)
                        channel_id += 1
                for section in self.apical:
                    for seg in section:
                        neuron.h.setdata_StochKv(seg.x, sec=section)
                        neuron.h.setRNG_StochKv(channel_id, self.cell_id.id)
                        channel_id += 1
            else:
                self.cell.re_init_rng()

    def get_psection(self, section_id: int | str) -> PSection:
        """Return a python section with the specified section id."""
        if isinstance(section_id, int):
            return self.psections[section_id]
        elif isinstance(section_id, str):
            return self.secname_to_psection[section_id]
        else:
            raise BluecellulabError(
                f"Section id must be an int or a str, not {type(section_id)}"
            )

    def make_passive(self) -> None:
        """Make the cell passive by deactivating all the active channels."""
        for section in self.all:
            mech_names = set()
            for seg in section:
                for mech in seg:
                    mech_names.add(mech.name())
            for mech_name in mech_names:
                if mech_name not in ["k_ion", "na_ion", "ca_ion", "pas",
                                     "ttx_ion"]:
                    neuron.h('uninsert %s' % mech_name, sec=section)
        self.is_made_passive = True

    def enable_ttx(self) -> None:
        """Add TTX to the environment (i.e. block the Na channels).

        Enable TTX by inserting TTXDynamicsSwitch and setting ttxo to
        1.0
        """
        if hasattr(public_hoc_cell(self.cell), 'enable_ttx'):
            public_hoc_cell(self.cell).enable_ttx()
        else:
            self._default_enable_ttx()

    def disable_ttx(self) -> None:
        """Remove TTX from the environment (i.e. unblock the Na channels).

        Disable TTX by inserting TTXDynamicsSwitch and setting ttxo to
        1e-14
        """
        if hasattr(public_hoc_cell(self.cell), 'disable_ttx'):
            public_hoc_cell(self.cell).disable_ttx()
        else:
            self._default_disable_ttx()

    def _default_enable_ttx(self) -> None:
        """Default enable_ttx implementation."""
        for section in self.all:
            if not neuron.h.ismembrane("TTXDynamicsSwitch"):
                section.insert('TTXDynamicsSwitch')
            section.ttxo_level_TTXDynamicsSwitch = 1.0

    def _default_disable_ttx(self) -> None:
        """Default disable_ttx implementation."""
        for section in self.all:
            if not neuron.h.ismembrane("TTXDynamicsSwitch"):
                section.insert('TTXDynamicsSwitch')
            section.ttxo_level_TTXDynamicsSwitch = 1e-14

    def area(self) -> float:
        """The total surface area of the cell."""
        area = 0.0
        for section in self.all:
            x_s = np.arange(1.0 / (2 * section.nseg), 1.0,
                            1.0 / (section.nseg))
            for x in x_s:
                area += neuron.h.area(x, sec=section)
            # for segment in section:
            #    area += neuron.h.area(segment.x, sec=section)
        return area

    def add_recording(self, var_name: str, dt: Optional[float] = None) -> None:
        """Add a recording to the cell.

        Args:
            var_name: Variable to be recorded.
            dt: Recording time step. If not provided, the recording step will
            default to the simulator's time step.
        """
        recording = neuron.h.Vector()
        if dt:
            # This float_epsilon stuff is some magic from M. Hines to make
            # the time points fall exactly on the dts
            recording.record(
                eval_neuron(var_name, self=self, neuron=neuron),
                self.get_precise_record_dt(dt),
            )
        else:
            recording.record(eval_neuron(var_name, self=self, neuron=neuron))
        self.recordings[var_name] = recording

    @staticmethod
    def get_precise_record_dt(dt: float) -> float:
        """Get a more precise record_dt to make time points faill on dts."""
        return (1.0 + neuron.h.float_epsilon) / (1.0 / dt)

    def add_recordings(self, var_names: list[str], dt: Optional[float] = None) -> None:
        """Add a list of recordings to the cell.

        Args:
            var_names: Variables to be recorded.
            dt: Recording time step. If not provided, the recording step will
            default to the simulator's time step.
        """
        for var_name in var_names:
            self.add_recording(var_name, dt)

    def add_ais_recording(self, dt: Optional[float] = None) -> None:
        """Adds recording to AIS."""
        self.add_recording("self.axonal[1](0.5)._ref_v", dt=dt)

    def add_voltage_recording(
        self, section: "neuron.h.Section", segx: float = 0.5, dt: Optional[float] = None
    ) -> None:
        """Add a voltage recording to a certain section at a given segment
        (segx).

        Args:
            section: Section to record from (Neuron section pointer).
            segx: Segment x coordinate. Specify a value between 0 and 1.
                  0 is typically the end closest to the soma, 1 is the distal end.
            dt: Recording time step. If not provided, the recording step will
                default to the simulator's time step.
        """
        var_name = f"neuron.h.{section.name()}({segx})._ref_v"
        self.add_recording(var_name, dt)

    def get_voltage_recording(
        self, section: "neuron.h.Section", segx: float = 0.5
    ) -> np.ndarray:
        """Get a voltage recording for a certain section at a given segment
        (segx).

        Args:
            section: Section to record from (Neuron section pointer).
            segx: Segment x coordinate. Specify a value between 0 and 1.
                  0 is typically the end closest to the soma, 1 is the distal end.

        Returns:
            A NumPy array containing the voltage recording values.

        Raises:
            BluecellulabError: If voltage recording was not added previously using add_voltage_recording.
        """
        recording_name = f"neuron.h.{section.name()}({segx})._ref_v"
        if recording_name in self.recordings:
            return self.get_recording(recording_name)
        else:
            raise BluecellulabError(
                f"get_voltage_recording: Voltage recording {recording_name}"
                " was not added previously using add_voltage_recording"
            )

    def add_allsections_voltagerecordings(self):
        """Add a voltage recording to every section of the cell."""
        all_sections = public_hoc_cell(self.cell).all
        for section in all_sections:
            self.add_voltage_recording(section, dt=self.record_dt)

    def get_allsections_voltagerecordings(self) -> dict[str, np.ndarray]:
        """Get all the voltage recordings from all the sections."""
        all_section_voltages = {}
        all_sections = public_hoc_cell(self.cell).all
        for section in all_sections:
            recording = self.get_voltage_recording(section)
            all_section_voltages[section.name()] = recording
        return all_section_voltages

    def get_recording(self, var_name: str) -> np.ndarray:
        """Get recorded values."""
        return np.array(self.recordings[var_name].to_python())

    def add_replay_synapse(self,
                           synapse_id: SynapseID,
                           syn_description: pd.Series,
                           connection_modifiers: dict,
                           condition_parameters: Conditions,
                           popids: tuple[int, int],
                           extracellular_calcium: float | None) -> None:
        """Add synapse based on the syn_description to the cell."""
        synapse = SynapseFactory.create_synapse(
            cell=self,
            syn_id=synapse_id,
            syn_description=syn_description,
            condition_parameters=condition_parameters,
            popids=popids,
            extracellular_calcium=extracellular_calcium,
            connection_modifiers=connection_modifiers)

        self.synapses[synapse_id] = synapse

        logger.debug(f'Added synapse to cell {self.cell_id.id}')

    def add_replay_delayed_weight(
        self, sid: tuple[str, int], delay: float, weight: float
    ) -> None:
        """Add a synaptic weight for sid that will be set with a time delay."""
        self.delayed_weights.put((delay, (sid, weight)))

    def pre_gids(self) -> list[int]:
        """Get the list of unique gids of cells that connect to this cell.

        Returns:
            A list of gids of cells that connect to this cell.
        """
        pre_gids = {self.synapses[syn_id].pre_gid for syn_id in self.synapses}
        return list(pre_gids)

    def pre_gid_synapse_ids(self, pre_gid: int) -> list[SynapseID]:
        """List of synapse_ids of synapses a cell uses to connect to this cell.

        Args:
            pre_gid: gid of the presynaptic cell.

        Returns:
            synapse_id's that connect the presynaptic cell with this cell.
        """
        syn_id_list = []
        for syn_id in self.synapses:
            if self.synapses[syn_id].pre_gid == pre_gid:
                syn_id_list.append(syn_id)
        return syn_id_list

    def create_netcon_spikedetector(self, target: HocObjectType, location: str, threshold: float = -30.0) -> HocObjectType:
        """Add and return a spikedetector.

        This is a NetCon that detects spike in the current cell, and that
        connects to target

        Args:
            target: target point process
            location: the spike detection location
            threshold: spike detection threshold

        Returns: Neuron netcon object

        Raises:
            ValueError: If the spike detection location is not 'soma' or 'AIS'.
        """
        if location == "soma":
            sec = public_hoc_cell(self.cell).soma[0]
            source = public_hoc_cell(self.cell).soma[0](1)._ref_v
        elif location == "AIS":
            sec = public_hoc_cell(self.cell).axon[1]
            source = public_hoc_cell(self.cell).axon[1](0.5)._ref_v
        else:
            raise ValueError("Spike detection location must be soma or AIS")
        netcon = neuron.h.NetCon(source, target, sec=sec)
        netcon.threshold = threshold
        return netcon

    def start_recording_spikes(self, target: HocObjectType, location: str, threshold: float = -30) -> None:
        """Start recording spikes in the current cell.

        Args:
            target: target point process
            location: the spike detection location
            threshold: spike detection threshold
        """
        nc = self.create_netcon_spikedetector(target, location, threshold)
        spike_vec = neuron.h.Vector()
        nc.record(spike_vec)
        self.recordings[f"spike_detector_{location}_{threshold}"] = spike_vec

    def get_recorded_spikes(self, location: str, threshold: float = -30) -> list[float]:
        """Get recorded spikes in the current cell.

        Args:
            location: the spike detection location
            threshold: spike detection threshold

        Returns: recorded spikes
        """
        result = self.recordings[f"spike_detector_{location}_{threshold}"]
        return result.to_python()

    def add_replay_minis(self,
                         synapse_id: SynapseID,
                         syn_description: pd.Series,
                         connection_modifiers: dict,
                         popids: tuple[int, int],
                         mini_frequencies: tuple[float | None, float | None]) -> None:
        """Add minis from the replay."""
        source_popid, target_popid = popids

        sid = synapse_id[1]

        weight = syn_description[SynapseProperty.G_SYNX]
        # numpy int to int
        post_sec_id = int(syn_description[SynapseProperty.POST_SECTION_ID])

        location = SynapseFactory.determine_synapse_location(
            syn_description, self
        )

        weight_scalar = connection_modifiers.get('Weight', 1.0)
        exc_mini_frequency, inh_mini_frequency = mini_frequencies \
            if mini_frequencies is not None else (None, None)

        synapse = self.synapses[synapse_id]

        # SpontMinis in sim config takes precedence of values in nodes file
        if 'SpontMinis' in connection_modifiers:
            spont_minis_rate = connection_modifiers['SpontMinis']
        elif synapse.mech_name in ["GluSynapse", "ProbAMPANMDA_EMS"]:
            spont_minis_rate = exc_mini_frequency
        else:
            spont_minis_rate = inh_mini_frequency

        if spont_minis_rate is not None and spont_minis_rate > 0:
            sec = self.get_psection(post_sec_id).hsection
            # add the *minis*: spontaneous synaptic events
            self.ips[synapse_id] = neuron.h.\
                InhPoissonStim(location, sec=sec)

            self.syn_mini_netcons[synapse_id] = neuron.h.\
                NetCon(self.ips[synapse_id], synapse.hsynapse, sec=sec)
            self.syn_mini_netcons[synapse_id].delay = 0.1
            self.syn_mini_netcons[synapse_id].weight[0] = weight * weight_scalar
            # set netcon type
            nc_param_name = f'nc_type_param_{synapse.hsynapse}'.split('[')[0]
            if hasattr(neuron.h, nc_param_name):
                nc_type_param = int(getattr(neuron.h, nc_param_name))
                # NC_SPONTMINI
                self.syn_mini_netcons[synapse_id].weight[nc_type_param] = 1

            rng_settings = RNGSettings.get_instance()
            if rng_settings.mode == 'Random123':
                seed2 = source_popid * 65536 + target_popid \
                    + rng_settings.minis_seed
                self.ips[synapse_id].setRNGs(
                    sid + 200,
                    self.cell_id.id + 250,
                    seed2 + 300,
                    sid + 200,
                    self.cell_id.id + 250,
                    seed2 + 350)
            else:
                exprng = neuron.h.Random()
                self.persistent.append(exprng)

                uniformrng = neuron.h.Random()
                self.persistent.append(uniformrng)

                base_seed = rng_settings.base_seed
                if rng_settings.mode == 'Compatibility':
                    exp_seed1 = sid * 100000 + 200
                    exp_seed2 = self.cell_id.id + 250 + base_seed + \
                        rng_settings.minis_seed
                    uniform_seed1 = sid * 100000 + 300
                    uniform_seed2 = self.cell_id.id + 250 + base_seed + \
                        rng_settings.minis_seed
                elif rng_settings.mode == "UpdatedMCell":
                    exp_seed1 = sid * 1000 + 200
                    exp_seed2 = source_popid * 16777216 + self.cell_id.id + 250 + \
                        base_seed + \
                        rng_settings.minis_seed
                    uniform_seed1 = sid * 1000 + 300
                    uniform_seed2 = source_popid * 16777216 + self.cell_id.id + 250 \
                        + base_seed + \
                        rng_settings.minis_seed
                else:
                    raise ValueError(
                        f"Cell: Unknown rng mode: {rng_settings.mode}")

                exprng.MCellRan4(exp_seed1, exp_seed2)
                exprng.negexp(1.0)

                uniformrng.MCellRan4(uniform_seed1, uniform_seed2)
                uniformrng.uniform(0.0, 1.0)

                self.ips[synapse_id].setRNGs(exprng, uniformrng)

            tbins_vec = neuron.h.Vector(1)
            tbins_vec.x[0] = 0.0
            rate_vec = neuron.h.Vector(1)
            rate_vec.x[0] = spont_minis_rate
            self.persistent.append(tbins_vec)
            self.persistent.append(rate_vec)
            self.ips[synapse_id].setTbins(tbins_vec)
            self.ips[synapse_id].setRate(rate_vec)

    def get_childrensections(self, parentsection: HocObjectType) -> list[HocObjectType]:
        """Get the children section of a neuron section."""
        number_children = neuron.h.SectionRef(sec=parentsection).nchild()
        children = []
        for index in range(int(number_children)):
            children.append(neuron.h.SectionRef(sec=self.soma).child[index])
        return children

    @staticmethod
    def get_parentsection(childsection: HocObjectType) -> HocObjectType:
        """Get the parent section of a neuron section."""
        return neuron.h.SectionRef(sec=childsection).parent

    def addAxialCurrentRecordings(self, section):
        """Record all the axial current flowing in and out of the section."""
        secname = neuron.h.secname(sec=section)
        self.add_recording(secname)
        for child in self.get_childrensections(section):
            self.add_recording(child)
        self.get_parentsection(section)

    def getAxialCurrentRecording(self, section):
        """Return the axial current recording."""
        secname = neuron.h.secname(sec=section)
        for child in self.get_childrensections(section):
            self.get_recording(secname)
            self.get_recording(child)

    def somatic_branches(self) -> None:
        """Show the index numbers."""
        nchild = neuron.h.SectionRef(sec=self.soma).nchild()
        for index in range(int(nchild)):
            secname = neuron.h.secname(sec=neuron.h.SectionRef(
                sec=self.soma).child[index])
            if "axon" not in secname:
                if "dend" in secname:
                    dendnumber = int(
                        secname.split("dend")[1].split("[")[1].split("]")[0])
                    secnumber = int(public_hoc_cell(self.cell).nSecAxonalOrig +
                                    public_hoc_cell(self.cell).nSecSoma + dendnumber)
                elif "apic" in secname:
                    apicnumber = int(secname.split(
                        "apic")[1].split("[")[1].split("]")[0])
                    secnumber = int(public_hoc_cell(self.cell).nSecAxonalOrig +
                                    public_hoc_cell(self.cell).nSecSoma +
                                    public_hoc_cell(self.cell).nSecBasal + apicnumber)
                    logger.info((apicnumber, secnumber))
                else:
                    raise BluecellulabError(
                        f"somaticbranches: No apic or dend found in section {secname}"
                    )

    @staticmethod
    @deprecated("Use bluecellulab.cell.section_distance.EuclideanSectionDistance instead.")
    def euclid_section_distance(
            hsection1=None,
            hsection2=None,
            location1=None,
            location2=None,
            projection=None):
        """Calculate euclidian distance between positions on two sections Uses
        bluecellulab.cell.section_distance.EuclideanSectionDistance.

        Parameters
        ----------

        hsection1 : hoc section
                    First section
        hsection2 : hoc section
                    Second section
        location1 : float
                    range x along hsection1
        location2 : float
                    range x along hsection2
        projection : string
                     planes to project on, e.g. 'xy'
        """
        dist = EuclideanSectionDistance()
        return dist(hsection1, hsection2, location1, location2, projection)

    def apical_trunk(self):
        """Return the apical trunk of the cell."""
        if len(self.apical) == 0:
            return []
        else:
            apicaltrunk = []
            max_diam_section = self.apical[0]
            while True:
                apicaltrunk.append(max_diam_section)

                children = [
                    neuron.h.SectionRef(sec=max_diam_section).child[index]
                    for index in range(int(neuron.h.SectionRef(
                        sec=max_diam_section).nchild()))]
                if len(children) == 0:
                    break
                maxdiam = 0
                for child in children:
                    if child.diam > maxdiam:
                        max_diam_section = child
                        maxdiam = child.diam
            return apicaltrunk

    def get_time(self) -> np.ndarray:
        """Get the time vector."""
        return self.get_recording('neuron.h._ref_t')

    def get_soma_voltage(self) -> np.ndarray:
        """Get a vector of the soma voltage."""
        return self.get_recording('self.soma(0.5)._ref_v')

    def get_ais_voltage(self) -> np.ndarray:
        """Get a vector of AIS voltage."""
        return self.get_recording('self.axonal[1](0.5)._ref_v')

    def getNumberOfSegments(self) -> int:
        """Get the number of segments in the cell."""
        return sum(section.nseg for section in self.all)

    def add_synapse_replay(
        self, stimulus: SynapseReplay, spike_threshold: float, spike_location: str
    ) -> None:
        """Adds the synapse spike replay to the cell if the synapse is
        connected to that cell."""
        if self.sonata_proxy is None:
            raise BluecellulabError("Cell: add_synapse_replay requires a sonata proxy.")
        synapse_spikes: dict = get_synapse_replay_spikes(stimulus.spike_file)
        for synapse_id, synapse in self.synapses.items():
            source_population = synapse.syn_description["source_population_name"]
            pre_gid = CellId(
                source_population, int(synapse.syn_description[SynapseProperty.PRE_GID])
            )
            if self.sonata_proxy.circuit_access.target_contains_cell(stimulus.source, pre_gid):
                if pre_gid.id in synapse_spikes:
                    spikes_of_interest = synapse_spikes[pre_gid.id]
                    # filter spikes of interest >=stimulus.delay, <=stimulus.duration
                    spikes_of_interest = spikes_of_interest[
                        (spikes_of_interest >= stimulus.delay)
                        & (spikes_of_interest <= stimulus.duration)
                    ]
                    connection = bluecellulab.Connection(
                        synapse,
                        pre_spiketrain=spikes_of_interest,
                        pre_cell=None,
                        stim_dt=self.record_dt,
                        spike_threshold=spike_threshold,
                        spike_location=spike_location,
                    )
                    logger.debug(
                        f"Added synapse replay from {pre_gid} to {self.cell_id.id}, {synapse_id}"
                    )

                    self.connections[synapse_id] = connection

    @property
    def info_dict(self):
        """Return a dictionary with all the information of this cell."""
        return {
            'synapses': {
                sid: synapse.info_dict for sid, synapse in self.synapses.items()
            },
            'connections': {
                sid: connection.info_dict for sid, connection in self.connections.items()
            }
        }

    def delete(self):
        """Delete the cell."""
        self.delete_plottable()
        if hasattr(self, 'cell') and self.cell is not None:
            if public_hoc_cell(self.cell) is not None and hasattr(public_hoc_cell(self.cell), 'clear'):
                public_hoc_cell(self.cell).clear()

            self.connections = None
            self.synapses = None

        if hasattr(self, 'recordings'):
            for recording in self.recordings:
                del recording

        if hasattr(self, 'persistent'):
            for persistent_object in self.persistent:
                del persistent_object

    def __del__(self):
        self.delete()
